# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A collection of `pathlib.Path`'s & file IO utilities.
"""

__all__ = []


import aiofiles
from pathlib import Path
from hashlib import sha3_256
from secrets import token_bytes

from ._typing import Typing as t
from ._constants import FILENAME_HASH_SLICE
from ._exceptions import Issue
from .asynchs import aos


# TODO @rmlibre: finish coverage tests
# fmt: off


def RootPath() -> Path:
    """
    Returns a `pathlib.Path` object pointing to this module's directory.
    """
    return Path(__file__).absolute().parent


def DatabasePath() -> Path:
    """
    Returns a `pathlib.Path` object pointing to the default directory
    where encrypted database files are saved.
    """
    return RootPath() / "db"


async def adeniable_filename(key: bytes, *, size: int = 8) -> str:
    """
    XORs `size` length bytes-type segments of a `key`, returning a
    variably forgeable, base-38 encoded hash from the condensed value.
    """
    from .generics.transform import axi_mix

    if 16 < size < 1:
        raise Issue.value_must("size", "be <= 16 and > 0")      # pragma: no cover
    filename = sha3_256(await axi_mix(key, size=size)).digest()
    return await t.ByteIO.abytes_to_filename(filename[FILENAME_HASH_SLICE])


def deniable_filename(key: bytes, *, size: int = 8) -> str:
    """
    XORs `size` length bytes-type segments of a `key`, returning a
    variably forgeable, base-38 encoded hash from the condensed value.
    """
    from .generics.transform import xi_mix

    if 16 < size < 1:
        raise Issue.value_must("size", "be <= 16 and > 0")      # pragma: no cover
    filename = sha3_256(xi_mix(key, size=size)).digest()
    return t.ByteIO.bytes_to_filename(filename[FILENAME_HASH_SLICE])


async def afind_salt_file(path: Path, *, key: bytes) -> Path:
    """
    Derives the filename to a cryptographic salt from a `key` & returns
    it joined to the provided `path` directory.
    """
    return path.absolute() / await adeniable_filename(key)


def find_salt_file(path: Path, *, key: bytes) -> Path:
    """
    Derives the filename to a cryptographic salt from a `key` & returns
    it joined to the provided `path` directory.
    """
    return path.absolute() / deniable_filename(key)


async def amake_salt_file(path: Path, *, salt: bytes = b"") -> None:
    """
    Saves a cryptographic `salt` contained within the file located at
    `path`. If no `salt` is provided, a new random 32-byte salt is used.
    """
    salt = salt if salt else token_bytes(32)
    async with aiofiles.open(path, "wb") as salt_file:
        await salt_file.write(salt)
    await aos.chmod(path, 0o000)


def make_salt_file(path: Path, *, salt: bytes = b"") -> None:
    """
    Saves a cryptographic `salt` contained within the file located at
    `path`. If no `salt` is provided, a new random 32-byte salt is used.
    """
    salt = salt if salt else token_bytes(32)
    path = Path(path)
    with path.open("wb") as salt_file:
        salt_file.write(salt)
    path.chmod(0o000)


async def aread_salt_file(path: t.PathStr) -> bytes:
    """
    Returns the cryptographic salt contained within the file located at
    `path`.
    """
    try:
        await aos.chmod(path, 0o700)
        async with aiofiles.open(path, "rb") as salt_file:
            salt = await salt_file.read()
            if len(salt) >= 32:
                return salt
            else:
                raise ValueError("len(salt) must be >= 32")     # pragma: no cover
    finally:
        await aos.chmod(path, 0o000)


def read_salt_file(path: t.PathStr) -> bytes:
    """
    Returns the cryptographic salt contained within the file located at
    `path`.
    """
    try:
        path = Path(path)
        path.chmod(0o700)
        with path.open("rb") as salt_file:
            salt = salt_file.read()
            if len(salt) >= 32:
                return salt
            else:
                raise ValueError("len(salt) must be >= 32")     # pragma: no cover
    finally:
        path.chmod(0o000)


async def aupdate_salt_file(path: t.PathStr, *, salt: bytes) -> None:
    """
    Replaces the cryptographic salt contained within the file located at
    `path` with the new `salt` value.
    """
    if len(salt) < 32:                                          # pragma: no cover
        raise ValueError("len(salt) must be >= 32")             # pragma: no cover
    try:                                                        # pragma: no cover
        await aos.chmod(path, 0o700)                            # pragma: no cover
        async with aiofiles.open(path, "wb") as salt_file:      # pragma: no cover
            await salt_file.write(salt)                         # pragma: no cover
    finally:                                                    # pragma: no cover
        await aos.chmod(path, 0o000)                            # pragma: no cover


def update_salt_file(path: t.PathStr, *, salt: bytes) -> None:
    """
    Replaces the cryptographic salt contained within the file located at
    `path` with the new `salt` value.
    """
    if len(salt) < 32:
        raise ValueError("len(salt) must be >= 32")             # pragma: no cover
    try:
        path = Path(path)
        path.chmod(0o700)
        with path.open("wb") as salt_file:
            salt_file.write(salt)
    finally:
        path.chmod(0o000)


async def adelete_salt_file(path: t.PathStr) -> None:
    """
    Deletes cryptographic salt contained within the file located at
    `path`.
    """
    await aos.chmod(path, 0o700)
    await aos.remove(path)


def delete_salt_file(path: t.PathStr) -> None:
    """
    Deletes cryptographic salt contained within the file located at
    `path`.
    """
    path = Path(path)
    path.chmod(0o700)
    path.unlink(path)


async def AsyncSecurePath(
    path: t.OptionalPathStr = None,
    *,
    key: t.Optional[bytes] = None,
    _admin: bool = False,
) -> Path:
    """
    Returns either a directory path where sensitive files are stored, or
    the file in that directory whose name is derived from `key`. The
    `_admin` flag is for internal package management of such files.
    """
    path = (Path(path).absolute() if path else DatabasePath()) / "secure"
    admin_path = path / "_admin"
    0 if path.is_dir() else await aos.mkdir(path)
    0 if admin_path.is_dir() else await aos.mkdir(admin_path)
    if key:
        fp = await afind_salt_file(admin_path if _admin else path, key=key)
        0 if fp.is_file() else await amake_salt_file(fp)
        return fp
    else:
        return path                                             # pragma: no cover


def SecurePath(
    path: t.OptionalPathStr = None,
    *,
    key: t.Optional[bytes] = None,
    _admin: bool = False,
) -> Path:
    """
    Returns either a directory path where sensitive files are stored, or
    the file in that directory whose name is derived from `key`. The
    `_admin` flag is for internal package management of such files.
    """
    path = (Path(path).absolute() if path else DatabasePath()) / "secure"
    admin_path = path / "_admin"
    0 if path.is_dir() else path.mkdir()
    0 if admin_path.is_dir() else admin_path.mkdir()
    if key:
        fp = find_salt_file(admin_path if _admin else path, key=key)
        0 if fp.is_file() else make_salt_file(fp)
        return fp
    else:
        return path                                             # pragma: no cover


# fmt: on


module_api = dict(
    AsyncSecurePath=AsyncSecurePath,
    DatabasePath=DatabasePath,
    Path=Path,
    RootPath=RootPath,
    SecurePath=SecurePath,
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    _afind_salt_file=afind_salt_file,
    _adelete_salt_file=adelete_salt_file,
    _amake_salt_file=amake_salt_file,
    _aread_salt_file=aread_salt_file,
    _aupdate_salt_file=aupdate_salt_file,
    _find_salt_file=find_salt_file,
    _delete_salt_file=delete_salt_file,
    _make_salt_file=make_salt_file,
    _read_salt_file=read_salt_file,
    _update_salt_file=update_salt_file,
    adeniable_filename=adeniable_filename,
    deniable_filename=deniable_filename,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Where data & key streams meet for confidentiality & authenticity.
"""

__all__ = ["StreamJunction"]


from aiootp._typing import Typing as t
from aiootp._constants import ENCRYPTION, DECRYPTION, ASYNC, SYNC
from aiootp._exceptions import Issue, KeyAADIssue
from aiootp.commons import FrozenInstance

from .stream_hmac import StreamHMAC


class StreamJunction(FrozenInstance):
    """
    A general definition for how the key & data streams are combined.
    """

    __slots__ = ()

    @classmethod
    def abytes_encipher(
        cls, data: t.AsyncOrSyncDatastream, *, shmac: StreamHMAC
    ) -> t.AsyncGenerator[bytes, None]:
        """
        A low-level function which returns an async generator that
        streams this package's ciphers.

        WARNING: `data` MUST produce plaintext in blocksize chunks or
        smaller per iteration or security MAY BE BROKEN by directly
        leaking plaintext. The cipher is designed for plaintext be
        padded using the `Padding` class.

        WARNING: The `finalize` or `afinalize` methods must be called on
        the `shmac` once all of the cipehrtext has been created /
        decrypted. Then the final SHMAC is available from the `aresult`
        & `result` methods. Authentication is attained by running the
        `(a)test_shmac` method. The `shmac` also has `(a)next_block_id`
        methods that can be used to authenticate unfinished streams of
        cipehrtext on the fly.
        """
        shmac._key_bundle._register_shmac(shmac)
        if shmac.mode != ENCRYPTION:
            raise Issue.must_set_value("shmac", ENCRYPTION)
        elif shmac._key_bundle._mode != ASYNC:
            raise KeyAADIssue.mode_isnt_correct(ASYNC)
        return cls.acombine_streams(data, shmac=shmac)

    @classmethod
    def bytes_encipher(
        cls, data: t.Datastream, *, shmac: StreamHMAC
    ) -> t.Generator[bytes, None, None]:
        """
        A low-level function which returns a sync generator that streams
        this package's ciphers.

        WARNING: `data` MUST produce plaintext in blocksize chunks or
        smaller per iteration or security MAY BE BROKEN by directly
        leaking plaintext. The cipher is designed for plaintext be
        padded using the `Padding` class.

        WARNING: The `finalize` or `afinalize` methods must be called on
        the `shmac` once all of the cipehrtext has been created /
        decrypted. Then the final SHMAC is available from the `aresult`
        & `result` methods. Authentication is attained by running the
        `(a)test_shmac` method. The `shmac` also has `(a)next_block_id`
        methods that can be used to authenticate unfinished streams of
        cipehrtext on the fly.
        """
        shmac._key_bundle._register_shmac(shmac)
        if shmac.mode != ENCRYPTION:
            raise Issue.must_set_value("shmac", ENCRYPTION)
        elif shmac._key_bundle._mode != SYNC:
            raise KeyAADIssue.mode_isnt_correct(SYNC)
        return cls.combine_streams(data, shmac=shmac)

    @classmethod
    def abytes_decipher(
        cls, data: t.AsyncOrSyncDatastream, *, shmac: StreamHMAC
    ) -> t.AsyncGenerator[bytes, None]:
        """
        A low-level function which returns an async generator that
        streams this package's ciphers.

        WARNING: The `finalize` or `afinalize` methods must be called on
        the `shmac` once all of the cipehrtext has been created /
        decrypted. Then the final SHMAC is available from the `aresult`
        & `result` methods. Authentication is attained by running the
        `(a)test_shmac` method. The `shmac` also has `(a)next_block_id`
        methods that can be used to authenticate unfinished streams of
        cipehrtext on the fly.
        """
        shmac._key_bundle._register_shmac(shmac)
        if shmac.mode != DECRYPTION:
            raise Issue.must_set_value("shmac", DECRYPTION)
        elif shmac._key_bundle._mode != ASYNC:
            raise KeyAADIssue.mode_isnt_correct(ASYNC)
        return cls.acombine_streams(data, shmac=shmac)

    @classmethod
    def bytes_decipher(
        cls, data: t.Datastream, *, shmac: StreamHMAC
    ) -> t.Generator[bytes, None, None]:
        """
        A low-level function which returns a sync generator that streams
        this package's ciphers.

        WARNING: The `finalize` or `afinalize` methods must be called on
        the `shmac` once all of the cipehrtext has been created /
        decrypted. Then the final SHMAC is available from the `aresult`
        & `result` methods. Authentication is attained by running the
        `(a)test_shmac` method. The `shmac` also has `(a)next_block_id`
        methods that can be used to authenticate unfinished streams of
        cipehrtext on the fly.
        """
        shmac._key_bundle._register_shmac(shmac)
        if shmac.mode != DECRYPTION:
            raise Issue.must_set_value("shmac", DECRYPTION)
        elif shmac._key_bundle._mode != SYNC:
            raise KeyAADIssue.mode_isnt_correct(SYNC)
        return cls.combine_streams(data, shmac=shmac)


module_api = dict(
    StreamJunction=t.add_type(StreamJunction),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Dependency inversion & documentation support for types relevant to
the `databases` subpackage.
"""

__all__ = ["AsyncDatabaseType", "DatabaseType"]


from .interface import Typing as t


@t.runtime_checkable
class AsyncDatabaseType(t.Protocol):
    async def aload_tags(self, *, silent: bool) -> t.Self:
        pass  # pragma: no cover

    async def aload_metatags(
        self, *, preload: bool, silent: bool
    ) -> t.Self:
        pass  # pragma: no cover

    async def aload_database(
        self,
        *,
        manifest: bool,
        silent: bool,
        preload: bool,
    ) -> t.Self:
        pass  # pragma: no cover

    async def afilename(self, tag: str) -> str:
        pass  # pragma: no cover

    async def aset_tag(
        self, tag: str, data: t.JSONSerializable, *, cache: bool
    ) -> t.Self:
        pass  # pragma: no cover

    async def aquery_tag(
        self, tag: str, *, silent: bool, cache: bool
    ) -> t.Union[bytes, t.JSONSerializable]:
        pass  # pragma: no cover

    async def apop_tag(
        self, tag: str, *, silent: bool
    ) -> t.Union[bytes, t.JSONSerializable]:
        pass  # pragma: no cover

    async def arollback_tag(self, tag: str, *, cache: bool) -> t.Self:
        pass  # pragma: no cover

    async def aclear_cache(self, *, metatags: bool) -> t.Self:
        pass  # pragma: no cover

    async def ametatag(
        self, tag: str, *, preload: bool, silent: bool
    ) -> t.Cls:
        pass  # pragma: no cover

    async def adelete_metatag(self, tag: str) -> t.Self:
        pass  # pragma: no cover

    async def adelete_database(self) -> None:
        pass  # pragma: no cover

    async def asave_tag(
        self, tag: str, *, admin: bool, drop_cache: bool
    ) -> t.Self:
        pass  # pragma: no cover


@t.runtime_checkable
class DatabaseType(t.Protocol):
    def load_tags(self, *, silent: bool) -> t.Self:
        pass  # pragma: no cover

    def load_metatags(self, *, preload: bool, silent: bool) -> t.Self:
        pass  # pragma: no cover

    def load_database(
        self,
        *,
        manifest: bool,
        silent: bool,
        preload: bool,
    ) -> t.Self:
        pass  # pragma: no cover

    def filename(self, tag: str) -> str:
        pass  # pragma: no cover

    def set_tag(
        self, tag: str, data: t.JSONSerializable, *, cache: bool
    ) -> t.Self:
        pass  # pragma: no cover

    def query_tag(
        self, tag: str, *, silent: bool, cache: bool
    ) -> t.Union[bytes, t.JSONSerializable]:
        pass  # pragma: no cover

    def pop_tag(
        self, tag: str, *, silent: bool
    ) -> t.Union[bytes, t.JSONSerializable]:
        pass  # pragma: no cover

    def rollback_tag(self, tag: str, *, cache: bool) -> t.Self:
        pass  # pragma: no cover

    def clear_cache(self, *, metatags: bool) -> t.Self:
        pass  # pragma: no cover

    def metatag(self, tag: str, *, preload: bool, silent: bool) -> t.Cls:
        pass  # pragma: no cover

    def delete_metatag(self, tag: str) -> t.Self:
        pass  # pragma: no cover

    def delete_database(self) -> None:
        pass  # pragma: no cover

    def save_tag(
        self, tag: str, *, admin: bool, drop_cache: bool
    ) -> t.Self:
        pass  # pragma: no cover


module_api = dict(
    AsyncDatabaseType=t.add_type(AsyncDatabaseType),
    DatabaseType=t.add_type(DatabaseType),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Dependency inversion & documentation support for types relevant to
the `keygens` subpackage.
"""

__all__ = [
    "AsymmetricKeyType",
    "DomainKDFType",
    "KeyExchangeProtocolType",
    "KeyExchangeType",
    "PublicKeyType",
    "SecretKeyType",
    "SignerType",
]


from .interface import Typing as t


@t.runtime_checkable
class DomainKDFType(t.Protocol):
    def copy(self) -> t.Cls:
        pass  # pragma: no cover

    async def aupdate(self, *data: bytes) -> t.Self:
        pass  # pragma: no cover

    def update(self, *data: bytes) -> t.Self:
        pass  # pragma: no cover

    async def asha3_256(self, *data: bytes, aad: bytes) -> bytes:
        pass  # pragma: no cover

    def sha3_256(self, *data: bytes, aad: bytes) -> bytes:
        pass  # pragma: no cover

    async def asha3_512(self, *data: bytes, aad: bytes) -> bytes:
        pass  # pragma: no cover

    def sha3_512(self, *data: bytes, aad: bytes) -> bytes:
        pass  # pragma: no cover

    async def ashake_128(
        self, *data: bytes, size: int, aad: bytes
    ) -> bytes:
        pass  # pragma: no cover

    def shake_128(self, *data: bytes, size: int, aad: bytes) -> bytes:
        pass  # pragma: no cover

    async def ashake_256(
        self, *data: bytes, size: int, aad: bytes
    ) -> bytes:
        pass  # pragma: no cover

    def shake_256(self, *data: bytes, size: int, aad: bytes) -> bytes:
        pass  # pragma: no cover


@t.runtime_checkable
class PublicKeyType(t.Protocol):
    def public_bytes(self, encoding: t.Any, format: t.Any) -> bytes:
        pass  # pragma: no cover

    def public_bytes_raw(self) -> bytes:
        pass  # pragma: no cover


@t.runtime_checkable
class SecretKeyType(t.Protocol):
    def private_bytes(self, encoding: t.Any, format: t.Any) -> bytes:
        pass  # pragma: no cover

    def private_bytes_raw(self) -> bytes:
        pass  # pragma: no cover

    def public_key(self) -> PublicKeyType:
        pass  # pragma: no cover


@t.runtime_checkable
class AsymmetricKeyType(t.Protocol):
    async def aimport_secret_key(
        self, secret_key: t.Union[bytes, SecretKeyType]
    ) -> t.Self:
        pass  # pragma: no cover

    def import_secret_key(
        self, secret_key: t.Union[bytes, SecretKeyType]
    ) -> t.Self:
        pass  # pragma: no cover

    async def aimport_public_key(
        self, public_key: t.Union[bytes, PublicKeyType]
    ) -> t.Self:
        pass  # pragma: no cover

    def import_public_key(
        self, public_key: t.Union[bytes, PublicKeyType]
    ) -> t.Self:
        pass  # pragma: no cover

    async def agenerate(self) -> t.Self:
        pass  # pragma: no cover

    def generate(self) -> t.Self:
        pass  # pragma: no cover

    def has_secret_key(self) -> bool:
        pass  # pragma: no cover

    def has_public_key(self) -> bool:
        pass  # pragma: no cover


@t.runtime_checkable
class SignerType(AsymmetricKeyType, t.Protocol):
    async def asign(self, data: bytes) -> bytes:
        pass  # pragma: no cover

    def sign(self, data: bytes) -> bytes:
        pass  # pragma: no cover

    async def averify(
        self,
        signature: bytes,
        data: bytes,
        *,
        public_key: t.Union[None, bytes, PublicKeyType],
    ) -> None:
        pass  # pragma: no cover

    def verify(
        self,
        signature: bytes,
        data: bytes,
        *,
        public_key: t.Union[None, bytes, PublicKeyType],
    ) -> None:
        pass  # pragma: no cover


@t.runtime_checkable
class KeyExchangeProtocolType(t.Protocol):
    async def asend(self, *keys: bytes) -> t.Union[t.Tuple[bytes], bytes]:
        pass  # pragma: no cover

    def send(self, *keys: bytes) -> t.Union[t.Tuple[bytes], bytes]:
        pass  # pragma: no cover

    async def areceive(self, *keys: bytes) -> DomainKDFType:
        pass  # pragma: no cover

    def receive(self, *keys: bytes) -> DomainKDFType:
        pass  # pragma: no cover


@t.runtime_checkable
class KeyExchangeType(AsymmetricKeyType, t.Protocol):
    async def aexchange(
        self, public_key: t.Union[PublicKeyType, bytes]
    ) -> bytes:
        pass  # pragma: no cover

    def exchange(self, public_key: t.Union[PublicKeyType, bytes]) -> bytes:
        pass  # pragma: no cover

    @classmethod
    def dh2_client(cls) -> KeyExchangeProtocolType:
        pass  # pragma: no cover

    def dh2_server(self) -> KeyExchangeProtocolType:
        pass  # pragma: no cover

    def dh3_client(self) -> KeyExchangeProtocolType:
        pass  # pragma: no cover

    def dh3_server(self) -> KeyExchangeProtocolType:
        pass  # pragma: no cover


module_api = dict(
    AsymmetricKeyType=t.add_type(AsymmetricKeyType),
    DomainKDFType=t.add_type(DomainKDFType),
    KeyExchangeProtocolType=t.add_type(KeyExchangeProtocolType),
    KeyExchangeType=t.add_type(KeyExchangeType),
    PublicKeyType=t.add_type(PublicKeyType),
    SecretKeyType=t.add_type(SecretKeyType),
    SignerType=t.add_type(SignerType),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A sub-package to better organize the package's constants.
"""

__all__ = []


import asyncio

from aiootp._typing import Typing as t


class NamespaceMapping(t.SimpleNamespace):
    def __init__(
        self, mapping: t.Mapping[str, t.Any] = {}, **kwargs: t.Any
    ) -> None:
        super().__init__(**kwargs)
        self.__dict__.update(mapping)

    def __getitem__(self, name: str) -> t.Any:
        return self.__dict__[name]

    def __setitem__(self, name: str, value: t.Any) -> None:
        self.__dict__[name] = value

    def __delitem__(self, name: str) -> None:
        del self.__dict__[name]

    def __contains__(self, name: str) -> bool:
        return name in self.__dict__

    def __len__(self) -> int:
        return len(self.__dict__)

    async def __aiter__(self) -> str:
        for name in self.__dict__:
            await asyncio.sleep(0)
            yield name

    def __iter__(self) -> str:
        yield from self.__dict__

    def keys(self) -> t.Iterable[t.Hashable]:
        yield from self.__dict__

    def values(self) -> t.Iterable[t.Any]:
        yield from self.__dict__.values()

    def items(self) -> t.Iterable[t.Tuple[t.Hashable, t.Any]]:
        yield from self.__dict__.items()

    def update(
        self, mapping: t.Mapping[str, t.Any] = {}, /, **kw: t.Any
    ) -> None:
        self.__dict__.update(mapping, **kw)


def collect_non_private_constants(
    mapping: t.Mapping[str, t.Any],
) -> t.Mapping[str, t.Any]:
    """
    Selectively moves uppercase, non-private declarations within a
    `mapping` to a returned dictionary.
    """
    return {
        name: value
        for name, value in mapping.items()
        if (name.isupper() and not name.startswith("_"))
    }


from .misc import *
from .datasets import *


modules = dict(misc=misc, datasets=datasets)


module_api = dict(
    NamespaceMapping=t.add_type(NamespaceMapping),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A bijective, keyed permutation type based on affine cipher
arithmetic of varying domain sizes.
"""

__all__ = ["AffinePermutation"]


from aiootp._typing import Typing as t
from aiootp._constants import BIG
from aiootp._exceptions import Issue
from aiootp.asynchs import asleep
from aiootp.commons import ConfigMap, FrozenInstance

from .affine_permutation_config import AffinePermutationConfig


class AffinePermutation(FrozenInstance):
    """
    Creates instances which perform invertible, bijective, keyed affine
    permutations on integers within a given byte-domain.
    --------
    WARNING: Alone, this permutation is insufficient to mask inputs, or
    -------- protect from key recovery upon viewing outputs.
     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    from aiootp._permutations.affine_permutation import AffinePermutation

    domain_size = 1
    key = token_bytes(AffinePermutation.key_size(domain_size))
    aff = AffinePermutation(key=key, config_id=domain_size)

    domain = tuple(range(256 ** domain_size))
    permutation = tuple(aff.permute(i) for i in domain)
    assert domain != permutation
    assert len(domain) == len(set(permutation))
    assert domain == tuple(aff.invert(i) for i in permutation)
    """

    __slots__ = (
        "_MAX",
        "_PRIME",
        "_additive_key",
        "config",
    )

    _configs = ConfigMap(
        mapping={
            config_id: AffinePermutationConfig(size=config_id)
            for config_id in [*range(1, 33), 64, 128, 192, 256]
        },
        config_type=AffinePermutationConfig,
    )

    @classmethod
    def key_size(cls, config_id: t.Hashable) -> int:
        """
        Returns the integer size of the keys an instance uses which is
        defined by the configuration indexed by the `config_id`.
        """
        return cls._configs[config_id].KEY_SIZE

    def _process_additive_key(self, key: bytes, size: int) -> int:
        """
        Returns the bytes-type `key` as an integer after it's assured to
        be the correct `size`, which for blinding purposes, must be
        twice the domain size.
        """
        if len(key) == size:
            return int.from_bytes(key, BIG)
        else:
            raise Issue.invalid_length("additive_key", size)

    def __init__(self, *, key: bytes, config_id: t.Hashable = 16) -> None:
        """
        Ensures the input keys are the correct size, stores them as
        integers, & loads the instance with the configuration indexed by
        the `config_id`.
        """
        self.config = c = self._configs[config_id]
        self._additive_key = self._process_additive_key(key, c.KEY_SIZE)
        self._MAX = c.MAX
        self._PRIME = c.PRIME

    async def apermute(self, value: int) -> int:
        """
        The conditional function which forms a bijective permutation on
        inputs within the domain by cycle walking until its output
        doesn't fall outside of the domain.
        """
        await asleep()
        return self.permute(value)

    def permute(self, value: int) -> int:
        """
        The conditional function which forms a bijective permutation on
        inputs within the domain by cycle walking until its output
        doesn't fall outside of the domain.
        """
        value = (
            self.config.MULTIPLICATIVE_KEY * value + self._additive_key
        ) % self._PRIME
        while value > self._MAX:
            value = (
                self.config.MULTIPLICATIVE_KEY * value + self._additive_key
            ) % self._PRIME
        return value

    async def ainvert(self, value: int) -> int:
        """
        Inverts the conditional function which forms a bijective
        permutation on inputs within the domain by cycle walking until
        its output doesn't fall outside of the domain.
        """
        await asleep()
        return self.invert(value)

    def invert(self, value: int) -> int:
        """
        Inverts the conditional function which forms a bijective
        permutation on inputs within the domain by cycle walking until
        its output doesn't fall outside of the domain.
        """
        value = (
            self.config.INVERSE_KEY * (value - self._additive_key)
        ) % self._PRIME
        while value > self._MAX:
            value = (
                self.config.INVERSE_KEY * (value - self._additive_key)
            ) % self._PRIME
        return value


module_api = dict(
    AffinePermutation=t.add_type(AffinePermutation),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A script to run the package signer over the build.
"""

import json
from pathlib import Path
from getpass import getpass

from aiootp import (
    __doc__,
    __license__,
    __package__,
    __version__,
    __author__,
)
from aiootp import PackageSigner, PackageVerifier


if __name__ == "__main__" and getpass(
    "sign package? (y/N) "
).lower().strip().startswith("y"):
    with Path("SIGNATURE.txt").open("r") as sig:
        scope = json.loads(sig.read())[PackageSigner._SCOPE]
        print("current version:", __version__)
        print(f"current build: {scope['build_number']}\n")

    signer = PackageSigner(
        package=__package__,
        version=__version__,
        author=__author__,
        license=__license__,
        description=__doc__,
        build_number=getpass("build number: "),
    )
    signer.connect_to_secure_database(
        username=getpass("database username: ").encode(),
        passphrase=getpass("database key: ").encode(),
        salt=getpass("database salt: ").encode(),
        path=getpass("secure directory: "),
    )

    if (
        getpass("is the signing key already saved on this device? (Y/n) ")
        .lower()
        .strip()
        .startswith("n")
    ):
        signer.update_signing_key(
            bytes.fromhex(getpass("signing key: ").strip())
        )

    while (
        getpass("update public credentials? (y/N) ")
        .lower()
        .strip()
        .startswith("y")
    ):
        signer.update_public_credentials(
            **{getpass("name: ").strip(): getpass("value: ")}
        )

    with Path("MANIFEST.in").open("r") as manifest:
        filename_sheet = manifest.read().strip().split("\n")

    for line in filename_sheet:
        line = line.strip()
        if "SIGNATURE" in line or not line.startswith("include"):
            continue
        filename = line.split(" ")[-1]
        with Path(filename).open("rb") as source_file:
            signer.add_file(filename, source_file.read())

    signer.sign_package()
    summary = signer.summarize()
    verifier = PackageVerifier(signer.signing_key.public_bytes, path="")
    verifier.verify_summary(summary)
    with Path("SIGNATURE.txt").open("w+") as attestation:
        attestation.write(json.dumps(summary, indent=4))



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Curve25519 interfaces & adapters to the `cryptography` package.
"""

__all__ = ["Ed25519", "X25519"]


from .shared_interface import *
from .ed25519 import *
from .double_diffie_hellman_client import *
from .double_diffie_hellman_server import *
from .triple_diffie_hellman_client import *
from .triple_diffie_hellman_server import *
from .x25519 import *


modules = dict(
    double_diffie_hellman_client=double_diffie_hellman_client,
    double_diffie_hellman_server=double_diffie_hellman_server,
    ed25519=ed25519,
    shared_interface=shared_interface,
    triple_diffie_hellman_client=triple_diffie_hellman_client,
    triple_diffie_hellman_server=triple_diffie_hellman_server,
    x25519=x25519,
)


module_api = dict(
    Ed25519=Ed25519,
    X25519=X25519,
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A configurable container interface for ciphertexts.
"""

__all__ = ["Ciphertext"]


from io import BytesIO

from aiootp._typing import Typing as t
from aiootp._exceptions import InvalidCiphertextSize
from aiootp.commons import OpenFrozenSlots


class Ciphertext(OpenFrozenSlots):
    """
    Efficiently stores bytes type ciphertext organized by instance
    attributes.
    """

    __slots__ = ("shmac", "salt", "iv", "ciphertext", "config")

    _MAPPED_ATTRIBUTES: t.Tuple[str] = ("shmac", "salt", "iv", "ciphertext")

    InvalidCiphertextSize: type = InvalidCiphertextSize

    def __init__(self, data: bytes, *, config: t.ConfigType) -> None:
        """
        Decomposes a blob of ciphertext `data` bytes into an organized
        instance where the `shmac` auth tag, `salt`, & `iv` randomizers &
        the body of `ciphertext` are queriable through dotted attribute
        lookup. `config_id` finds a specific cipher's ciphertext format
        within the class' mapping of configurations.
        """
        self.config = config
        self._ensure_valid_size(len(data))
        data = BytesIO(data)
        self.shmac = data.read(config.SHMAC_BYTES)
        self.salt = data.read(config.SALT_BYTES)
        self.iv = data.read(config.IV_BYTES)
        self.ciphertext = data.read()

    def __iter__(self) -> t.Generator[str, None, None]:
        yield from self._MAPPED_ATTRIBUTES

    def _ensure_valid_size(self, data_length: int) -> None:
        """
        Accepts the integer length of a blob of ciphertext & throws
        `InvalidCiphertextSize` if an impossible value is detected.
        """
        size = data_length - self.config.HEADER_BYTES
        if size <= 0 or size % self.config.BLOCKSIZE:
            raise InvalidCiphertextSize(data_length)


module_api = dict(
    Ciphertext=t.add_type(Ciphertext),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
An interface to Ed25519 signing & verifying.
"""

__all__ = ["Ed25519"]


from aiootp._typing import Typing as t
from aiootp.asynchs import asleep

from .shared_interface import Base25519


class Ed25519(Base25519):
    """
    This class is used to create stateful objects that simplify usage of
    the cryptography library's ed25519 protocol.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    from aiootp import Ed25519

    # In a land, long ago ->
    alices_key = Ed25519().generate()
    internet.send(alices_key.public_bytes)

    # Alice wants to sign a document so that Bob can prove she wrote it.
    # So, Alice sends her public key bytes of the key she wants to
    # associate with her identity, the document & the signature ->
    document = b"DesignDocument.cad"
    signed_document = alices_key.sign(document)
    message = {
        "document": document,
        "signature": signed_document,
        "public_key": alices_key.public_bytes,
    }
    internet.send(message)

    # In a land far away ->
    alices_message = internet.receive()

    # Bob sees the message from Alice! Bob already knows Alice's public
    # key & she has reason believe it is genuinely hers. She'll then
    # verify the signed document ->
    assert alices_message["public_key"] == alices_public_key
    alice_verifier = Ed25519().import_public_key(alices_public_key)
    alice_verifier.verify(
        alices_message["signature"], alices_message["document"]
    )
    internet.send(b"Beautiful work, Alice! Thanks ^u^")

    # The verification didn't throw an exception! So, Bob knows the file
    # was signed by Alice.
    """

    __slots__ = ("_public_key", "_secret_key")

    InvalidSignature: type = t.InvalidSignature
    PublicKey: type = t.Ed25519PublicKey
    SecretKey: type = t.Ed25519PrivateKey

    async def asign(self, data: bytes, /) -> bytes:
        """
        Signs some bytes `data` with the instance's secret key.
        """
        await asleep()
        return self.secret_key.sign(data)

    def sign(self, data: bytes, /) -> bytes:
        """
        Signs some bytes `data` with the instance's secret key.
        """
        return self.secret_key.sign(data)

    async def averify(
        self,
        /,
        signature: bytes,
        data: bytes,
        *,
        public_key: t.Union[None, bytes, t.Ed25519PublicKey] = None,
    ) -> None:
        """
        Receives a `signature` to verify data with the instance's
        public key. If the `public_key` keyword-only argument is
        used, then that key is used instead of the instance key to run
        the verification.
        """
        await asleep()
        self.verify(signature=signature, data=data, public_key=public_key)

    def verify(
        self,
        /,
        signature: bytes,
        data: bytes,
        *,
        public_key: t.Union[None, bytes, t.Ed25519PublicKey] = None,
    ) -> None:
        """
        Receives a `signature` to verify data with the instance's
        public key. If the `public_key` keyword-only argument is
        used, then that key is used instead of the instance key to run
        the verification.
        """
        if public_key is None:
            public_key = self.public_key
        else:
            public_key = self._process_public_key(public_key)
        public_key.verify(signature, data)


module_api = dict(
    Ed25519=t.add_type(Ed25519),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
General definition for the StreamHMAC interface.
"""

__all__ = ["StreamHMAC"]


from aiootp._typing import Typing as t
from aiootp._constants import DEFAULT_AAD, BIG
from aiootp._constants import ENCRYPTION, DECRYPTION
from aiootp._exceptions import Issue, SHMACIssue, ValidationIncomplete
from aiootp._exceptions import InvalidBlockID, InvalidSHMAC
from aiootp.asynchs import asleep
from aiootp.generics import bytes_are_equal

from .key_bundle import KeyAADBundle


class StreamHMAC:
    """
    This class is used as an inline validator for ciphertext streams as
    they are being created & decrypted. It's designed for full context
    commitment, salt misuse-reuse resistance, & RUP security.
    """

    __slots__ = (
        "_aupdate",
        "_avalidated_transform",
        "_current_digest",
        "_is_finalized",
        "_key_bundle",
        "_previous_digest",
        "_mac",
        "_permutation",
        "_mode",
        "_result",
        "_result_is_ready",
        "_update",
        "_validated_transform",
        "config",
    )

    InvalidBlockID = InvalidBlockID
    InvalidSHMAC = InvalidSHMAC

    def __init__(self, key_bundle: KeyAADBundle) -> None:
        """
        Begins a stateful hash object that's used to calculate a keyed-
        message authentication code referred to as a shmac, as well as
        block IDs to validate ciphertext streams that have not yet been
        completed.
        """
        if not issubclass(key_bundle.__class__, KeyAADBundle):
            raise Issue.value_must_be_type("key_bundle", KeyAADBundle)
        self._mode = None
        self._is_finalized = False
        self._result_is_ready = False
        self._register_key_bundle(key_bundle)
        self._initialize_session_state()
        self._update = self._placeholder_update  # Don't allow updates
        self._aupdate = self._aplaceholder_update  # unless mode is set

    def _register_key_bundle(self, key_bundle: KeyAADBundle) -> None:
        """
        Registers the `KeyAADBundle` object & necessary values, which
        will be tied to the instance for a single run of the cipher.
        Reusing the same instance or `key_bundle` for multiple cipher
        calls is NOT SAFE, & is disallowed by this registration.
        """
        key_bundle._mode.validate()
        self._key_bundle = key_bundle
        self.config = key_bundle.config
        self._mac = key_bundle._shmac_kdf

    def _initialize_session_state(self) -> None:
        """
        Prepares the state for the current call session.
        """
        config = self.config
        primer_key = self._mac.digest(config.PRIMER_KEY_BYTES)
        self._previous_digest = self._current_digest = primer_key[
            config.FIRST_DIGEST_SLICE
        ]
        self._permutation = config.Permutation(
            key=primer_key[config.PERMUTATION_KEY_SLICE],
            config_id=config.PERMUTATION_CONFIG_ID,
        )

    @property
    def mode(self) -> str:
        """
        Returns the mode which the instance was instructed to be in
        after `_for_encryption` or `_for_decryption` are called.
        """
        return self._mode

    @property
    def result(self) -> bytes:
        """
        Returns the instance's final authentication tag result. Raises
        `PermissionError` if the instance hasn't been finalized.
        """
        if not self._result_is_ready:
            raise ValidationIncomplete
        return self._result

    def _for_encryption(self) -> t.Self:
        """
        Instructs the SHMAC instance to prepare itself for validating
        ciphertext while encrypting.
        """
        if self._mode:
            raise Issue.value_already_set("shmac", self._mode)
        elif self._key_bundle._iv_given_by_user:
            raise SHMACIssue.invalid_iv_usage()
        self._mode = ENCRYPTION
        self._update = self._update_mac
        self._aupdate = self._aupdate_mac
        self._validated_transform = self._encipher_then_hash
        self._avalidated_transform = self._aencipher_then_hash
        return self

    def _for_decryption(self) -> t.Self:
        """
        Instructs the SHMAC instance to prepare itself for validating
        ciphertext while decrypting.
        """
        if self._mode:
            raise Issue.value_already_set("shmac", self._mode)
        elif not self._key_bundle._iv_given_by_user:
            raise SHMACIssue.invalid_iv_usage()
        self._mode = DECRYPTION
        self._update = self._update_mac
        self._aupdate = self._aupdate_mac
        self._validated_transform = self._hash_then_decipher
        self._avalidated_transform = self._ahash_then_decipher
        return self

    async def _aplaceholder_update(self, *_: t.Any, **__: t.Any) -> None:
        """
        This method is overwritten with the propper functionality when a
        cipher mode is declared with either the `_for_encryption` or
        `_for_decryption` methods. This interface helps ensure correct
        usage of the object.
        """
        raise SHMACIssue.no_cipher_mode_declared()

    def _placeholder_update(self, *_: t.Any, **__: t.Any) -> None:
        """
        This method is overwritten with the propper functionality when a
        cipher mode is declared with either the `_for_encryption` or
        `_for_decryption` methods. This interface helps ensure correct
        usage of the object.
        """
        raise SHMACIssue.no_cipher_mode_declared()

    async def _aupdate_mac(self, ciphertext_block: bytes) -> t.Self:
        """
        This method is called automatically when an instance is passed
        into the low-level `(a)bytes_encipher` / `(a)bytes_decipher`
        generators.
        """
        await asleep()
        mac = self._mac
        self._previous_digest = self._current_digest
        mac.update(ciphertext_block)
        self._current_digest = mac.digest(self.config.SHMAC_BLOCKSIZE)
        return self

    def _update_mac(self, ciphertext_block: bytes) -> t.Self:
        """
        This method is called automatically when an instance is passed
        into the low-level `(a)bytes_encipher` / `(a)bytes_decipher`
        generators.
        """
        mac = self._mac
        self._previous_digest = self._current_digest
        mac.update(ciphertext_block)
        self._current_digest = mac.digest(self.config.SHMAC_BLOCKSIZE)
        return self

    def _test_block_id_size(self, size: int) -> None:
        """
        Raises errors if the requested `block_id` `size` is not within
        the allowed bounds.
        """
        if size < self.config.MIN_BLOCK_ID_BYTES:
            raise SHMACIssue.block_id_is_too_small(
                size, self.config.MIN_BLOCK_ID_BYTES
            )
        elif size > self.config.MAX_BLOCK_ID_BYTES:
            raise SHMACIssue.block_id_is_too_big(
                size, self.config.MAX_BLOCK_ID_BYTES
            )

    def _get_block_id_mac(self) -> None:
        """
        Returns a correct mac digest considering that during encryption
        the ciphertext is generated before the block id is generated, &
        it must be checked by an instance during decryption before it's
        decrypted.
        """
        if self._mode == ENCRYPTION:
            return self._previous_digest
        else:
            return self._current_digest

    async def anext_block_id(
        self,
        next_block: bytes,
        *,
        size: t.Optional[int] = None,
        aad: bytes = DEFAULT_AAD,
        _join: t.Callable[..., bytes] = b"".join,
    ) -> bytes:
        """
        Returns a `size`-byte block id derived from the current state
        & the supplied `next_block` chunk of ciphertext. These block
        ids can be used to detect out-of-order messages, as well as
        ciphertext forgeries, without altering the internal state. This
        allows for robust decryption of ciphertext streams, mitigates
        adversarial attempts to crash communication channels, & allows
        release of plaintexts without waiting for the stream to end.

        Additional `aad` for each block may be specified if desired,
        however the `aad` passed into this method DO NOT alter the
        keystream. It's only used to create block ids that authenticate
        the associated data.
        """
        await asleep()
        size = size if size else self.config.BLOCK_ID_BYTES
        self._test_block_id_size(size)
        mac = self.config.BLOCK_ID_KDF_CONFIG.factory()
        payload = (
            self._get_block_id_mac(),
            size.to_bytes(1, BIG),
            len(aad).to_bytes(8, BIG),
            aad,
            len(next_block).to_bytes(8, BIG),
            next_block,
        )
        mac.update(_join(payload))
        return mac.digest(size)

    def next_block_id(
        self,
        next_block: bytes,
        *,
        size: t.Optional[int] = None,
        aad: bytes = DEFAULT_AAD,
        _join: t.Callable[..., bytes] = b"".join,
    ) -> bytes:
        """
        Returns a `size`-byte block id derived from the current state
        & the supplied `next_block` chunk of ciphertext. These block
        ids can be used to detect out-of-order messages, as well as
        ciphertext forgeries, without altering the internal state. This
        allows for robust decryption of ciphertext streams & mitigates
        adversarial attempts to crash communication channels, & allows
        release of plaintexts without waiting for the stream to end.

        Additional `aad` for each block may be specified if desired,
        however the `aad` passed into this method DO NOT alter the
        keystream. It's only used to create block ids that authenticate
        the associated data.
        """
        size = size if size else self.config.BLOCK_ID_BYTES
        self._test_block_id_size(size)
        mac = self.config.BLOCK_ID_KDF_CONFIG.factory()
        payload = (
            self._get_block_id_mac(),
            size.to_bytes(1, BIG),
            len(aad).to_bytes(8, BIG),
            aad,
            len(next_block).to_bytes(8, BIG),
            next_block,
        )
        mac.update(_join(payload))
        return mac.digest(size)

    async def _aset_final_result(self) -> None:
        """
        Sets the instance's final result with a SHMAC of its state. This
        signals the end of a stream of data that can be validated with
        the current instance.
        """
        await asleep()
        self._result = self._mac.digest(
            self.config.SHMAC_DOUBLE_BLOCKSIZE  # fmt: skip
        )[self.config.SHMAC_RESULT_SLICE]

    def _set_final_result(self) -> None:
        """
        Sets the instance's final result with a SHMAC of its state. This
        signals the end of a stream of data that can be validated with
        the current instance.
        """
        self._result = self._mac.digest(
            self.config.SHMAC_DOUBLE_BLOCKSIZE  # fmt: skip
        )[self.config.SHMAC_RESULT_SLICE]

    async def afinalize(self) -> bytes:
        """
        Sets the instance's final result with a SHMAC of its state. This
        signals the end of a stream of data that can be validated with
        the current instance. Returns the authentication tag.
        """
        if self._is_finalized:
            raise SHMACIssue.already_finalized()
        self._is_finalized = True
        await self._aset_final_result()
        self._result_is_ready = True
        del self._mac
        del self._previous_digest
        del self._current_digest
        return self._result

    def finalize(self) -> bytes:
        """
        Sets the instance's final result with a SHMAC of its state. This
        signals the end of a stream of data that can be validated with
        the current instance. Returns the authentication tag.
        """
        if self._is_finalized:
            raise SHMACIssue.already_finalized()
        self._is_finalized = True
        self._set_final_result()
        self._result_is_ready = True
        del self._mac
        del self._previous_digest
        del self._current_digest
        return self._result

    async def atest_next_block_id(
        self,
        untrusted_block_id: bytes,
        next_block: bytes,
        aad: bytes = DEFAULT_AAD,
    ) -> None:
        """
        Does a timing-safe comparison of a supplied `untrusted_block_id`
        with a derived block id of the supplied `next_block` chunk of
        ciphertext. Raises `InvalidBlockID` if the untrusted block ID is
        invalid. These block id checks can detect out of order messages,
        or ciphertext forgeries, without altering the internal state.
        This allows for robust decryption of ciphertext streams,
        mitigates adversarial attempts to crash a communication channel,
        & allows release of plaintexts without waiting for the stream to
        end.

        The `aad` for each block may be specified to authenticate the
        associated data with the block.
        """
        if untrusted_block_id.__class__ is not bytes:
            raise Issue.value_must_be_type("untrusted_block_id", bytes)
        size = len(untrusted_block_id)
        block_id = await self.anext_block_id(next_block, size=size, aad=aad)
        if not bytes_are_equal(untrusted_block_id, block_id):
            raise SHMACIssue.invalid_block_id()

    def test_next_block_id(
        self,
        untrusted_block_id: bytes,
        next_block: bytes,
        aad: bytes = DEFAULT_AAD,
    ) -> None:
        """
        Does a timing-safe comparison of a supplied `untrusted_block_id`
        with a derived block id of the supplied `next_block` chunk of
        ciphertext. Raises `InvalidBlockID` if the untrusted block ID is
        invalid. These block id checks can detect out of order messages,
        or ciphertext forgeries, without altering the internal state.
        This allows for robust decryption of ciphertext streams,
        mitigates adversarial attempts to crash a communication channel,
        & allows release of plaintexts without waiting for the stream to
        end.

        The `aad` for each block may be specified to authenticate the
        associated data with the block.
        """
        if untrusted_block_id.__class__ is not bytes:
            raise Issue.value_must_be_type("untrusted_block_id", bytes)
        size = len(untrusted_block_id)
        block_id = self.next_block_id(next_block, size=size, aad=aad)
        if not bytes_are_equal(untrusted_block_id, block_id):
            raise SHMACIssue.invalid_block_id()

    async def atest_shmac(self, untrusted_shmac: bytes) -> None:
        """
        Does a time-safe comparison of a supplied `untrusted_shmac`
        with the instance's final result shmac. Raises `InvalidSHMAC` if
        the shmac doesn't match.
        """
        if untrusted_shmac.__class__ is not bytes:
            raise Issue.value_must_be_type("untrusted_shmac", bytes)
        elif not bytes_are_equal(untrusted_shmac, self.result):
            raise SHMACIssue.invalid_shmac()

    def test_shmac(self, untrusted_shmac: bytes) -> None:
        """
        Does a time-safe comparison of a supplied `untrusted_shmac`
        with the instance's final result shmac. Raises `InvalidSHMAC` if
        the shmac doesn't match.
        """
        if untrusted_shmac.__class__ is not bytes:
            raise Issue.value_must_be_type("untrusted_shmac", bytes)
        elif not bytes_are_equal(untrusted_shmac, self.result):
            raise SHMACIssue.invalid_shmac()


module_api = dict(
    StreamHMAC=t.add_type(StreamHMAC),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Passcrypt configuration logic & constants.
"""

__all__ = ["PasscryptConfig", "passcrypt_spec"]


from aiootp._typing import Typing as t
from aiootp._constants import INT_BYTES, BIG
from aiootp._constants import NANOSECONDS, SAFE_TIMESTAMP_BYTES, EPOCH_NS
from aiootp._exceptions import PasscryptIssue, Metadata
from aiootp.asynchs import Clock
from aiootp.commons import Config
from aiootp.generics import canonical_pack


class PasscryptConfig(Config):
    """
    Specifies the configuration for `Passcrypt`.
    """

    __slots__ = (
        "CORES_BYTES",
        "CORES_RESOURCE_SAFETY_RANGE",
        "CORES_SLICE",
        "CPU_BYTES",
        "CPU_RESOURCE_SAFETY_RANGE",
        "CPU_SLICE",
        "CPU_TO_DIGEST_PAIRS_PER_ROW_RATIO",
        "DEFAULT_CORES",
        "DEFAULT_CPU",
        "DEFAULT_MB",
        "DEFAULT_SCHEMA_SALT_SIZE",
        "EPOCH_NS",
        "HEADER_BYTES",
        "HEADER_SLICE",
        "MAX_CORES",
        "MAX_CPU",
        "MAX_MB",
        "MAX_SALT_SIZE",
        "MB_BYTES",
        "MB_RESOURCE_SAFETY_RANGE",
        "MB_SLICE",
        "MIN_CORES",
        "MIN_CPU",
        "MIN_MB",
        "MIN_PASSPHRASE_BYTES",
        "MIN_SALT_SIZE",
        "MIN_SCHEMA_BYTES",
        "MIN_TAG_SIZE",
        "PACKED_METADATA",
        "PASSCRYPT_PAD",
        "SALT_SIZE_BYTES",
        "SALT_SIZE_SLICE",
        "TIME_UNIT",
        "TIMESTAMP_BYTES",
        "TIMESTAMP_SLICE",
        "clock",
    )

    slots_types: t.Mapping[str, type] = dict(
        CONFIG_ID=t.Hashable,
        CORES_BYTES=int,
        CORES_RESOURCE_SAFETY_RANGE=range,
        CORES_SLICE=slice,
        CPU_BYTES=int,
        CPU_RESOURCE_SAFETY_RANGE=range,
        CPU_SLICE=slice,
        CPU_TO_DIGEST_PAIRS_PER_ROW_RATIO=int,
        DEFAULT_CORES=int,
        DEFAULT_CPU=int,
        DEFAULT_MB=int,
        DEFAULT_SCHEMA_SALT_SIZE=int,
        EPOCH_NS=int,
        HEADER_BYTES=int,
        HEADER_SLICE=slice,
        MAX_CORES=int,
        MAX_CPU=int,
        MAX_MB=int,
        MAX_SALT_SIZE=int,
        MB_BYTES=int,
        MB_RESOURCE_SAFETY_RANGE=range,
        MB_SLICE=slice,
        MIN_CORES=int,
        MIN_CPU=int,
        MIN_MB=int,
        MIN_PASSPHRASE_BYTES=int,
        MIN_SALT_SIZE=int,
        MIN_SCHEMA_BYTES=int,
        MIN_TAG_SIZE=int,
        PACKED_METADATA=bytes,
        PASSCRYPT_PAD=bytes,
        SALT_SIZE_BYTES=int,
        SALT_SIZE_SLICE=slice,
        TIME_UNIT=str,
        TIMESTAMP_BYTES=int,
        TIMESTAMP_SLICE=slice,
        clock=t.ClockType,
    )

    def __init__(
        self,
        *,
        config_id: t.Hashable,
        min_passphrase_bytes: int,
        passcrypt_pad: bytes,
        epoch_ns: int = EPOCH_NS,
    ) -> None:
        self.MIN_CORES = 1
        self.MAX_CORES = 256
        self.DEFAULT_CORES = 4
        self.CORES_BYTES = 1
        self.MIN_CPU = 1
        self.MAX_CPU = 256
        self.DEFAULT_CPU = 1
        self.CPU_TO_DIGEST_PAIRS_PER_ROW_RATIO = 2
        self.CPU_BYTES = 1
        self.MIN_MB = 1
        self.MAX_MB = 256**3
        self.DEFAULT_MB = 64
        self.MB_BYTES = 3
        self.MIN_SALT_SIZE = 4
        self.MAX_SALT_SIZE = 256
        self.DEFAULT_SCHEMA_SALT_SIZE = 8
        self.SALT_SIZE_BYTES = 1
        self.MIN_TAG_SIZE = 16
        self.TIME_UNIT = NANOSECONDS
        self.TIMESTAMP_BYTES = SAFE_TIMESTAMP_BYTES
        self.CONFIG_ID = config_id
        self.EPOCH_NS = epoch_ns
        self.MIN_PASSPHRASE_BYTES = min_passphrase_bytes
        self.PASSCRYPT_PAD = passcrypt_pad
        self._initialize_dynamic_values()
        self._construct_metadata_constant()

    def _initialize_dynamic_values(self) -> None:
        """
        Uses the specified constants to build dependent configuration
        constants & objects.
        """
        self.HEADER_BYTES = (
            self.TIMESTAMP_BYTES
            + self.MB_BYTES
            + self.CPU_BYTES
            + self.CORES_BYTES
            + self.SALT_SIZE_BYTES
        )
        self.HEADER_SLICE = slice(0, self.HEADER_BYTES, 1)
        self.TIMESTAMP_SLICE = slice(0, self.TIMESTAMP_BYTES, 1)
        self.MB_SLICE = slice(
            self.TIMESTAMP_BYTES, self.TIMESTAMP_BYTES + self.MB_BYTES, 1
        )
        self.MB_RESOURCE_SAFETY_RANGE = range(self.MIN_MB, 512)
        self.CPU_SLICE = slice(
            self.TIMESTAMP_BYTES + self.MB_BYTES,
            self.TIMESTAMP_BYTES + self.MB_BYTES + self.CPU_BYTES,
            1,
        )
        self.CPU_RESOURCE_SAFETY_RANGE = range(self.MIN_CPU, 33)
        self.CORES_SLICE = slice(
            self.TIMESTAMP_BYTES + self.MB_BYTES + self.CPU_BYTES,
            self.TIMESTAMP_BYTES
            + self.MB_BYTES
            + self.CPU_BYTES
            + self.CORES_BYTES,
            1,
        )
        self.CORES_RESOURCE_SAFETY_RANGE = range(self.MIN_CORES, 9)
        self.SALT_SIZE_SLICE = slice(
            self.TIMESTAMP_BYTES
            + self.MB_BYTES
            + self.CPU_BYTES
            + self.CORES_BYTES,
            self.HEADER_BYTES,
            1,
        )
        self.MIN_SCHEMA_BYTES = (
            self.HEADER_BYTES + self.MIN_SALT_SIZE + self.MIN_TAG_SIZE
        )
        self.clock = Clock(self.TIME_UNIT, epoch=self.EPOCH_NS)

    def _construct_metadata_constant(self) -> None:
        """
        Causes hashes to be distinct for distinct configurations.

        IMPORTANT FOR SECURITY.
        See:
        https://eprint.iacr.org/2016/292.pdf
        https://csrc.nist.gov/csrc/media/Events/2023/third-workshop-on-
            block-cipher-modes-of-operation/documents/accepted-papers/
            Flexible%20Authenticated%20Encryption.pdf

        DO NOT OVERRIDE TO PROVIDE ITER-OP.
        """
        self.PACKED_METADATA = canonical_pack(
            self.CONFIG_ID,
            str(self.CORES_RESOURCE_SAFETY_RANGE).encode(),
            str(self.CORES_SLICE).encode(),
            str(self.CPU_RESOURCE_SAFETY_RANGE).encode(),
            str(self.CPU_SLICE).encode(),
            self.CPU_TO_DIGEST_PAIRS_PER_ROW_RATIO.to_bytes(INT_BYTES, BIG),
            self.DEFAULT_CORES.to_bytes(INT_BYTES, BIG),
            self.DEFAULT_CPU.to_bytes(INT_BYTES, BIG),
            self.DEFAULT_MB.to_bytes(INT_BYTES, BIG),
            self.DEFAULT_SCHEMA_SALT_SIZE.to_bytes(INT_BYTES, BIG),
            self.EPOCH_NS.to_bytes(16, BIG),
            str(self.HEADER_SLICE).encode(),
            self.MAX_CORES.to_bytes(INT_BYTES, BIG),
            self.MAX_CPU.to_bytes(INT_BYTES, BIG),
            self.MAX_MB.to_bytes(INT_BYTES, BIG),
            self.MAX_SALT_SIZE.to_bytes(INT_BYTES, BIG),
            self.MB_BYTES.to_bytes(INT_BYTES, BIG),
            str(self.MB_RESOURCE_SAFETY_RANGE).encode(),
            str(self.MB_SLICE).encode(),
            self.MIN_CORES.to_bytes(INT_BYTES, BIG),
            self.MIN_CPU.to_bytes(INT_BYTES, BIG),
            self.MIN_MB.to_bytes(INT_BYTES, BIG),
            self.MIN_PASSPHRASE_BYTES.to_bytes(INT_BYTES, BIG),
            self.MIN_SALT_SIZE.to_bytes(INT_BYTES, BIG),
            self.MIN_SCHEMA_BYTES.to_bytes(INT_BYTES, BIG),
            self.MIN_TAG_SIZE.to_bytes(INT_BYTES, BIG),
            self.PASSCRYPT_PAD,
            str(self.SALT_SIZE_SLICE).encode(),
            self.TIME_UNIT.encode(),
            str(self.TIMESTAMP_SLICE).encode(),
            int_bytes=1,
        )

    def is_passphrase(self, passphrase: bytes) -> bool:
        return (
            passphrase.__class__ is bytes
            and len(passphrase) >= self.MIN_PASSPHRASE_BYTES
        )

    def is_salt(self, salt: bytes) -> bool:
        length_limits = range(self.MIN_SALT_SIZE, self.MAX_SALT_SIZE + 1)
        return (salt.__class__ is bytes) and (len(salt) in length_limits)

    def is_aad(self, aad: bytes) -> bool:
        return aad.__class__ is bytes

    def is_mb(self, mb: int) -> bool:
        mb_limits = range(self.MIN_MB, self.MAX_MB + 1)
        return (mb.__class__ is int) and (mb in mb_limits)

    def is_cpu(self, cpu: int) -> bool:
        cpu_limits = range(self.MIN_CPU, self.MAX_CPU + 1)
        return (cpu.__class__ is int) and (cpu in cpu_limits)

    def is_cores(self, cores: int) -> bool:
        cores_limits = range(self.MIN_CORES, self.MAX_CORES + 1)
        return (cores.__class__ is int) and (cores in cores_limits)

    def is_tag_size(self, tag_size: int) -> bool:
        tag_size_is_int = tag_size.__class__ is int
        return (tag_size_is_int) and (tag_size >= self.MIN_TAG_SIZE)

    def is_salt_size(self, salt_size: int) -> bool:
        salt_size_is_int = salt_size.__class__ is int
        salt_size_limits = range(self.MIN_SALT_SIZE, self.MAX_SALT_SIZE + 1)
        return salt_size_is_int and (salt_size in salt_size_limits)

    def validate_inputs(
        self, passphrase: bytes, salt: bytes, aad: bytes
    ) -> None:
        """
        Makes sure `passphrase`, `salt`, & `aad` are to specification.
        Throws `ValueError` or `TypeError` accordingly if not.
        """
        if not self.is_passphrase(passphrase):
            raise PasscryptIssue.improper_passphrase(Metadata(passphrase))
        elif not self.is_salt(salt):
            raise PasscryptIssue.improper_salt(Metadata(salt))
        elif not self.is_aad(aad):
            raise PasscryptIssue.improper_aad()

    def validate_settings(
        self,
        mb: int,
        cpu: int,
        cores: int,
        tag_size: int,
        salt_size: int,
    ) -> None:
        """
        Ensures the `mb`, `cpu`, `cores`, `tag_size`, & `salt_size`
        values are to specification. Throws `ValueError` or `TypeError`
        accordingly if not.
        """
        if not self.is_mb(mb):
            raise PasscryptIssue.invalid_mb(mb)
        elif not self.is_cpu(cpu):
            raise PasscryptIssue.invalid_cpu(cpu)
        elif not self.is_cores(cores):
            raise PasscryptIssue.invalid_cores(cores)
        elif not self.is_tag_size(tag_size):
            raise PasscryptIssue.invalid_tag_size(tag_size)
        elif not self.is_salt_size(salt_size):
            raise PasscryptIssue.invalid_salt_size(salt_size)


passcrypt_spec = PasscryptConfig(
    config_id=b"Passcrypt", min_passphrase_bytes=12, passcrypt_pad=b"\xf2"
)


module_api = dict(
    PasscryptConfig=t.add_type(PasscryptConfig),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    passcrypt_spec=passcrypt_spec,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
An interface for hashing & signing software packages.
"""

__all__ = ["PackageSigner", "PackageSignerFiles", "PackageSignerScope"]


import json
from hashlib import sha384

from aiootp._typing import Typing as t
from aiootp._constants import DAYS
from aiootp._constants import CHECKSUM, CHECKSUMS, PUBLIC_CREDENTIALS
from aiootp._constants import SCOPE, SIGNATURE, SIGNING_KEY, VERSIONS
from aiootp._exceptions import PackageSignerIssue
from aiootp.asynchs import Clock
from aiootp.generics import Domains, canonical_pack
from aiootp.databases import Database

from .curve_25519 import Ed25519


class PackageSignerScope(t.OpenNamespace):
    """
    Stores user-defined scope values of a package signing session.
    """

    def __init__(
        self,
        *,
        package: str,
        version: str,
        date: int,
        **extras: t.Dict[str, t.JSONSerializable],
    ) -> None:
        self.__dict__.update(extras)
        self.package = package
        self.version = version
        self.date = date


class PackageSignerFiles(t.OpenNamespace):
    """
    Stores the filename, hashing object key-value pairs of a package
    signing session.
    """


class PackageSigner:
    """
    Provides an intuitive API for users to sign their own packages.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    import json
    from getpass import getpass
    from aiootp import PackageSigner

    signer = PackageSigner(
        package=__package__,
        version=__version__,
        author=__author__,
        license=__license__,
    )
    signer.connect_to_secure_database(
        passphrase=getpass("database passphrase: "),
        salt=getpass("database salt: "),
        path=getpass("secure directory: "),
    )

    with open("MANIFEST.in", "r") as manifest:
        filename_sheet = manifest.read().split("\n")

    for line in filename_sheet:
        if not line.startswith("include"):
            continue
        filename = line.strip().split(" ")[-1]
        with open(filename, "rb") as source_file:
            signer.add_file(filename, source_file.read())

    signer.sign_package()
    signer.db.save_database()
    package_signature_summary = signer.summarize()
    """

    __slots__ = ("_db", "_scope", "files")

    _Database: type = Database
    _Hasher: t.Callable[[bytes], t.HasherType] = sha384
    _Signer: type = Ed25519

    _CHECKSUM: str = CHECKSUM
    _CHECKSUMS: str = CHECKSUMS
    _CLASS: str = "PackageSigner"
    _PUBLIC_CREDENTIALS: str = PUBLIC_CREDENTIALS
    _SCOPE: str = SCOPE
    _SIGNATURE: str = SIGNATURE
    _SIGNING_KEY: str = SIGNING_KEY
    _VERSIONS: str = VERSIONS

    _clock: t.ClockType = Clock(DAYS, epoch=0)

    InvalidSignature: type = PackageSignerIssue.InvalidSignature

    @classmethod
    def _database_template(cls) -> t.Dict[str, t.Union[str, t.JSONObject]]:
        """
        Returns the default instance package database values for
        initializing new databases.
        """
        return {
            cls._SIGNING_KEY: "",
            cls._VERSIONS: {},
            cls._PUBLIC_CREDENTIALS: {},
        }

    @classmethod
    def generate_signing_key(cls) -> t.SignerType:
        """
        Generates a new `Ed25519` secret signing key object.
        """
        return cls._Signer().generate()

    def __init__(
        self,
        package: str,
        version: str,
        date: t.Optional[int] = None,
        **scopes: t.JSONSerializable,
    ) -> None:
        """
        Sets the instance's package scope attributes & default file
        checksums container.
        """
        self.files = PackageSignerFiles()
        self._scope = PackageSignerScope(
            package=package,
            version=version,
            date=date if date else self._clock.time(),
            **scopes,
        )

    def __repr__(self) -> str:
        """
        Displays the instance's declared scope.
        """
        cls = self.__class__.__qualname__
        return str(f"{cls}({self._scope})")

    @property
    def _public_credentials(self) -> t.JSONObject:
        """
        Returns public credentials from the instance's secure database.
        """
        return self.db[self._scope.package][self._PUBLIC_CREDENTIALS]

    @property
    def signing_key(self) -> t.SignerType:
        """
        Returns the package's secret signing key from the instance's
        encrypted database in an `Ed25519` object.
        """
        package = self._scope.package
        aad = canonical_pack(Domains.PACKAGE_SIGNER, package.encode())
        encrypted_key = self.db[package][self._SIGNING_KEY]
        if not encrypted_key:
            raise PackageSignerIssue.SigningKeyNotSet
        key = self.db.read_token(encrypted_key, aad=aad)
        return self._Signer().import_secret_key(key)

    @property
    def db(self) -> t.DatabaseType:
        """
        Returns the instance's database object, or alerts the user to
        connect to a secure database if it isn't yet set.
        """
        try:
            return self._db
        except AttributeError:
            raise PackageSignerIssue.DatabaseNotConnected from None

    @property
    def _checksums(self) -> t.Dict[str, str]:
        """
        Returns the instance's package filenames & their hexdigests in
        a JSON ready dictionary.
        """
        return {
            filename: hasher.hexdigest()
            for filename, hasher in sorted(self.files.items())
        }

    @property
    def _summary(self) -> t.JSONObject:
        """
        Collects the instance's package file checksums, names, public
        credentials, version scopes & the package's public signing key
        into a json-ready dictionary.
        """
        return {
            self._CHECKSUMS: dict(self._checksums.items()),
            self._PUBLIC_CREDENTIALS: dict(
                sorted(self._public_credentials.items())
            ),
            self._SCOPE: dict(sorted(self._scope.items())),
            self._SIGNING_KEY: self.signing_key.public_bytes.hex(),
        }

    @property
    def _checksum(self) -> bytes:
        """
        Returns the digest of the current package summary.
        """
        return self._Hasher(json.dumps(self._summary).encode()).digest()

    @property
    def _signature(self) -> bytes:
        """
        Returns the stored package summary signature.
        """
        try:
            versions = self.db[self._scope.package][self._VERSIONS]
            return bytes.fromhex(versions[self._scope.version])
        except KeyError as error:
            raise PackageSignerIssue.PackageNotSigned from error

    def connect_to_secure_database(
        self,
        *secret_credentials: bytes,
        username: bytes,
        passphrase: bytes,
        salt: bytes = b"",
        path: t.OptionalPathStr = None,
        **passcrypt_settings: t.Union[int, t.ConfigType],
    ) -> t.Self:
        """
        Opens an encrypted database connection using the Passcrypt
        passphrase-based key derivation function, a `passphrase` & any
        available additional credentials a user may have. If a database
        doesn't already exist, then a new one is created with default
        values.
        """
        self._db = self._Database.generate_profile(
            self._CLASS.encode(),
            *secret_credentials,
            username=username,
            passphrase=passphrase,
            salt=salt,
            path=path,
            **passcrypt_settings,
        )
        try:
            self.db.query_tag(self._scope.package, cache=True)
        except LookupError:
            self.db[self._scope.package] = self._database_template()
        return self

    def update_scope(self, **scopes: t.JSONSerializable) -> t.Self:
        """
        Updates the package scopes to qualify the package signature of
        the current package version within the instance.
        """
        self._scope.__dict__.update(scopes)
        return self

    def update_public_credentials(
        self, **credentials: t.JSONSerializable
    ) -> t.Self:
        """
        Updates the public credentials to be associated with the package
        signature & stores them in the instance's database cache. The
        database must be saved separately to save them to disk.
        """
        package = self._scope.package
        self.db[package][self._PUBLIC_CREDENTIALS].update(credentials)
        return self

    def update_signing_key(
        self, signing_key: t.Union[bytes, t.SecretKeyType, t.SignerType]
    ) -> t.Self:
        """
        Updates the package's secret signing key as an encrypted token
        within the instance's database cache. The database must be saved
        separately to save the encrypted signing key to disk.
        """
        if signing_key.__class__ is not self._Signer:
            signing_key = self._Signer().import_secret_key(signing_key)
        package = self._scope.package
        aad = canonical_pack(Domains.PACKAGE_SIGNER, package.encode())
        self.db[package][self._SIGNING_KEY] = self.db.make_token(
            signing_key.secret_bytes, aad=aad
        ).decode()
        return self

    def add_file(self, filename: str, file_data: bytes) -> t.Self:
        """
        Stores a `filename` & the hash object of the file's bytes type
        contents in the instance's `files` attribute mapping.
        """
        self.files[filename] = self._Hasher(file_data)
        return self

    def sign_package(self) -> t.Self:
        """
        Signs the package summary checksum & stores it in the instance's
        secure database cache. The database must be saved separately to
        save the signature to disk.
        """
        checksum = self._checksum
        with self.db as db:
            db[self._scope.package][self._VERSIONS].update(
                {self._scope.version: self.signing_key.sign(checksum).hex()}
            )
        return self

    def summarize(self) -> t.Dict[str, t.JSONSerializable]:
        """
        Assures the stored package checksum signature matches the
        current checksum of the package summary. If valid, the summary
        is returned along with the package checksum & its signature in
        a dictionary.
        """
        checksum = self._checksum
        signature = self._signature
        signing_key = self.signing_key
        try:
            signing_key.verify(signature, checksum)
        except self.InvalidSignature as error:
            raise PackageSignerIssue.unsynced_package_signature() from error
        return {
            self._CHECKSUM: checksum.hex(),
            **self._summary,
            self._SIGNATURE: signature.hex(),
        }


module_api = dict(
    PackageSigner=t.add_type(PackageSigner),
    PackageSignerFiles=t.add_type(PackageSignerFiles),
    PackageSignerScope=t.add_type(PackageSignerScope),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A controller for toggling debug mode.
"""

__all__ = ["DebugControl"]


import warnings

from ._typing import Typing as t


class DebugControl:
    """
    WARNING: Enabling debugging reveals potentially sensitive values,
    -------- such as cryptographic keys, in object repr's that are
    omitted by default. Also turns on asyncio's debugging.
    """

    __slots__ = ()

    _DEBUG_MODE: bool = False

    _switches: t.List[t.Callable[[bool], t.Any]] = []

    @classmethod
    def is_debugging(cls, /) -> bool:
        return cls._DEBUG_MODE

    @classmethod
    def enable_debugging(cls, /, *, silence_warnings: bool = False) -> None:
        """
        WARNING: This will reveal potentially sensitive values, such as
        cryptographic keys, in object repr's that are omitted by default.
        """
        if not silence_warnings:
            warnings.warn(cls.__doc__)  # pragma: no cover
        if not cls._DEBUG_MODE:
            cls._DEBUG_MODE = True
            for toggle in cls._switches:
                toggle()

    @classmethod
    def disable_debugging(cls, /) -> None:
        if cls._DEBUG_MODE:
            cls._DEBUG_MODE = False
            for toggle in cls._switches:
                toggle()


module_api = dict(
    DebugControl=DebugControl,
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Composed passcrypt hash logic, formatting, & parsing types.
"""

__all__ = ["PasscryptResources", "PasscryptHash"]


from io import BytesIO

from aiootp._typing import Typing as t
from aiootp._constants import BIG, MB, CPU, CORES
from aiootp._exceptions import PasscryptIssue, raise_exception
from aiootp.commons import FrozenSlots, OpenFrozenSlots

from .config import passcrypt_spec


class PasscryptResources(OpenFrozenSlots):
    """
    Efficiently stores the resource values located in the header of a
    `Passcrypt` hash.
    """

    __slots__ = ("mb", "cpu", "cores")

    def __init__(self, mb: int, cpu: int, cores: int) -> None:
        self.mb = mb
        self.cpu = cpu
        self.cores = cores


class PasscryptHash(FrozenSlots):
    """
    Efficiently stores `Passcrypt` session values to be encoded into &
    safely decoded from formatted `Passcrypt` hashes.
    """

    __slots__ = ("timestamp", "mb", "cpu", "cores", "salt", "tag", "config")

    _MAPPED_ATTRIBUTES: t.Tuple[str] = (
        "timestamp",
        "mb",
        "cpu",
        "cores",
        "salt",
        "tag",
    )

    def __init__(
        self,
        *,
        timestamp: t.Optional[bytes] = None,
        mb: t.Optional[int] = None,
        cpu: t.Optional[int] = None,
        cores: t.Optional[int] = None,
        salt: t.Optional[bytes] = None,
        tag: t.Optional[bytes] = None,
        config: t.ConfigType = passcrypt_spec,
    ) -> None:
        """
        Populates the instance state from the provided session values
        which are composable into a `Passcrypt` hash.
        """
        object.__setattr__(self, "config", config)
        if any((timestamp, mb, cpu, cores, salt, tag)):
            config.validate_settings(
                mb=mb,
                cpu=cpu,
                cores=cores,
                tag_size=len(tag),
                salt_size=len(salt),
            )
            self.timestamp = timestamp
            self.mb = mb
            self.cpu = cpu
            self.cores = cores
            self.salt = salt
            self.tag = tag

    def __iter__(self) -> t.Generator[str, None, None]:
        yield from self._MAPPED_ATTRIBUTES

    @property
    def salt_size(self) -> t.Optional[int]:
        """
        Returns the length of the `salt` value stored in the instance
        state. If the `salt` has not been set, returns `None`.
        """
        salt = getattr(self, "salt", None)
        return len(salt) if salt else None

    @property
    def tag_size(self) -> t.Optional[int]:
        """
        Returns the length of the `tag` value stored in the instance
        state. If the `tag` has not been set, returns `None`.
        """
        tag = getattr(self, "tag", None)
        return len(tag) if tag else None

    def import_hash(self, passcrypt_hash: bytes) -> t.Self:
        """
        Populates the instance state from the decoded values represented
        in the bytes-type `passcrypt_hash`. These hashes contain the
        inputs & parameters of a `Passcrypt` session.
        """
        config = self.config
        to_int = int.from_bytes
        read = BytesIO(passcrypt_hash).read
        self.__init__(
            timestamp=read(config.TIMESTAMP_BYTES),
            mb=to_int(read(config.MB_BYTES), BIG) + 1,
            cpu=to_int(read(config.CPU_BYTES), BIG) + 1,
            cores=to_int(read(config.CORES_BYTES), BIG) + 1,
            salt=read(to_int(read(config.SALT_SIZE_BYTES), BIG) + 1),
            tag=read(),
        )
        return self

    def export_hash(self) -> bytes:
        """
        Returns the composed `Passcrypt` hash from the instance state.
        """
        config = self.config
        passcrypt_hash = (
            self.timestamp,
            (self.mb - 1).to_bytes(config.MB_BYTES, BIG),
            (self.cpu - 1).to_bytes(config.CPU_BYTES, BIG),
            (self.cores - 1).to_bytes(config.CORES_BYTES, BIG),
            (self.salt_size - 1).to_bytes(config.SALT_SIZE_BYTES, BIG),
            self.salt,
            self.tag,
        )
        return b"".join(passcrypt_hash)

    def in_allowed_ranges(
        self, mb_allowed: range, cpu_allowed: range, cores_allowed: range
    ) -> bool:
        """
        Procs a `ResourceWarning` exception if any of the range objects
        passed into the method do not contain the value which is set for
        its specified difficulty setting.

         _____________________________________
        |                                     |
        |            Usage Example:           |
        |_____________________________________|

        from aiootp import Passcrypt

        allowed_resource_consumption = dict(
            mb_allowed=range(16, 256),  # Less than 256 MiB allowed
            cpu_allowed=range(2, 8),    # Less than 8 complexity allowed
            cores_allowed=range(1, 5),  # Less than 5 processes allowed
        )

        try:
            Passcrypt.verify(hashed_pw, pw, **allowed_resource_consumption)
        except ResourceWarning as danger:
            admin.log(danger)
            hard_limits_exceeded = (
                danger.requested_resources.mb > 512
                or danger.requested_resources.cpu > 11
                or danger.requested_resources.cores > 8
            )
            below_security_guidelines = (
                danger.requested_resources.mb < 16
                or danger.requested_resources.cpu < 2
            )
            if hard_limits_exceeded:
                raise danger
            elif below_security_guidelines:
                raise PermissionError("Minimum hash difficulty unmet.")
            Passcrypt.verify(hashed_pw, pw)
        """
        proc = raise_exception
        exc = PasscryptIssue.untrusted_resource_consumption
        header = PasscryptResources(self.mb, self.cpu, self.cores)
        self.mb in mb_allowed or proc(exc(MB, header))
        self.cpu in cpu_allowed or proc(exc(CPU, header))
        self.cores in cores_allowed or proc(exc(CORES, header))
        return True


module_api = dict(
    PasscryptHash=t.add_type(PasscryptHash),
    PasscryptResources=t.add_type(PasscryptResources),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Definitions for mapping classes that expect subclasses to define
instance attributes in __slots__.
"""

__all__ = ["FrozenSlots", "OpenFrozenSlots", "OpenSlots", "Slots"]


import asyncio
from os import linesep as sep

from aiootp._typing import Typing as t
from aiootp._constants import OMITTED
from aiootp._debug_control import DebugControl
from aiootp._exceptions import Issue
from aiootp._gentools import collate


class Slots:
    """
    A base class which allow subclasses to create very efficient
    instances, with explicitly declared attributes in their `__slots__`.

    Masked repr.
    Mutable instance.

     _____________________________________
    |                                     |
    |   Stability of Assignment Styles:   |
    |_____________________________________|


    from aiootp.commons.slots import Slots

    class HybridNamespace(Slots):
        __slots__ = ("attr", "__dict__")

    hybrid = HybridNamespace()

    ✔ hybrid.attr = "value"                # supported
    ✔ hybrid["attr"] = "value"             # supported
    ✔ setattr(hybrid, "attr", "value")     # supported

    ❌ hybrid.__dict__["attr"] = "value"    # unsupported


    # See: https://github.com/rmlibre/aiootp/pull/11
    """

    __slots__ = ()

    _MAPPED_ATTRIBUTES: t.Tuple[str] = ()
    _UNMAPPED_ATTRIBUTES: t.Tuple[str] = (
        "__dict__",
        "_MAPPED_ATTRIBUTES",
        "_UNMAPPED_ATTRIBUTES",
        "_is_mapped_attribute",
        "keys",
        "values",
        "items",
        "update",
    )

    def __init_subclass__(cls, /, *a: t.Any, **kw: t.Any) -> None:
        """
        Brings slots declarations from subclasses up the class hierarchy.
        """
        super().__init_subclass__(*a, **kw)
        # Preserve original declaration order & enforce uniqueness
        # fmt: off
        cls.__slots__ = tuple({
            name: None
            for subcls in cls.__mro__
            for name in getattr(subcls, "__slots__", ())
        })
        # fmt: on
        cls._slots_set = frozenset(cls.__slots__)

    def __init__(
        self, mapping: t.Mapping[t.Hashable, t.Any] = {}, /, **kw: t.Any
    ) -> None:
        """
        Maps user-defined kwargs to instance attributes. If a subclass
        defines a `__slots__`, then only the names declared within can
        be admitted to the instance. Unless `"__dict__"` is also added.
        Defining classes with slots improves a system's memory efficiency,
        especially when many instances are created. Having an instance
        dictionary offers flexibility, though the interplay with slots
        can cause problems. This initializer avoids setting the names
        declared in `__slots__` within the a potential instance dict.

        This is a general & flexible initializer. Subclasses can improve
        their performance & documentation by being more specific, ie:

        def __init__(self, a: int, b: int) -> None:
            self.a = a
            self.b = b
        """
        self.update(mapping, **kw)

    def __dir__(self, /) -> t.List[t.Hashable]:
        """
        Returns the instance directory.
        """
        return list(
            set(object.__dir__(self))
            .difference(self._UNMAPPED_ATTRIBUTES)
            .union(self._MAPPED_ATTRIBUTES)
        )

    def __bool__(self, /) -> bool:
        """
        If the instance is empty then return False, otherwise True.
        """
        return any(1 for name in self)

    def __len__(self, /) -> int:
        """
        Returns the number of elements in the instance.
        """
        return sum(1 for name in self)

    def __contains__(self, name: t.Hashable, /) -> bool:
        """
        Returns a bool of `name`'s membership in the instance.
        """
        in_slots = name in self._slots_set
        in_dict = name in getattr(self, "__dict__", ())
        return (in_slots and hasattr(self, name)) or in_dict

    def __setattr__(self, name: str, value: t.Any, /) -> None:
        """
        Delegates attr assignment logic to the `__getitem__` interface.
        This allows more consistency for subclasses with custom behavior
        since changes can be applied in one place. That interface also
        allows for non-string attribute names, better supporting those
        subclasses which define `__dict__`.

         _____________________________________
        |                                     |
        |   Stability of Assignment Styles:   |
        |_____________________________________|


        from aiootp.commons.slots import Slots

        class HybridNamespace(Slots):
            __slots__ = ("attr", "__dict__")

        hybrid = HybridNamespace()

        ✔ hybrid.attr = "value"                # supported
        ✔ hybrid["attr"] = "value"             # supported
        ✔ setattr(hybrid, "attr", "value")     # supported

        ❌ hybrid.__dict__["attr"] = "value"    # unsupported


        # See: https://github.com/rmlibre/aiootp/pull/11
        """
        self[name] = value

    def __delattr__(self, name: str, /) -> None:
        """
        Delegates attr deletion logic to the `__detitem__` interface.
        This allows more consistency for subclasses with custom behavior
        since changes can be applied in one place. That interface also
        allows for non-string attribute names, better supporting those
        subclasses which define `__dict__`.
        """
        del self[name]

    def __setitem__(self, name: str, value: t.Any, /) -> None:
        """
        Transforms bracket item assignment into dotted assignment on the
        instance.
        """
        if name.__class__ is str:
            object.__setattr__(self, name, value)
        else:
            self.__dict__[name] = value

    def __getitem__(self, name: str, /) -> t.Any:
        """
        Transforms bracket lookup into dotted access on the instance.
        """
        if name.__class__ is str:
            return getattr(self, name)
        else:
            return self.__dict__[name]

    def __delitem__(self, name: str, /) -> None:
        """
        Deletes the item `name` from the instance.
        """
        if name.__class__ is str:
            object.__delattr__(self, name)
        else:
            del self.__dict__[name]

    def __repr__(self, /, *, mask: bool = True) -> str:
        """
        Pretty displays the instance & its attributes.
        """
        cls = self.__class__
        if mask and not DebugControl.is_debugging():
            show = lambda value: f"{OMITTED}{value.__class__}"
        else:
            show = lambda value: repr(value)
        start = f"{sep}    " if self else ""
        on = f",{sep}    " if self else ", "
        body = on.join(
            f"{name}={show(value)}"
            for name, value in self.items()
            if str(name)[0] != "_"
        )
        end = f",{sep}" if self else ""
        return f"{cls.__qualname__}({start}{body}{end})"

    def _is_mapped_attribute(self, name: str, /) -> bool:
        """
        Allows the class to define criteria which include an instance
        attribute within the mapping unpacking interface.
        """
        mapped = name in self._MAPPED_ATTRIBUTES
        unmapped = name in self._UNMAPPED_ATTRIBUTES
        return mapped or not unmapped

    async def __aiter__(self, /) -> t.AsyncGenerator[t.Any, None]:
        """
        Unpacks instance variable names with with async iteration.
        """
        for name in self:
            await asyncio.sleep(0)
            yield name

    def __iter__(self, /) -> t.Generator[t.Any, None, None]:
        """
        Unpacks instance variable names with with sync iteration.
        """
        for name in collate(self.__slots__, getattr(self, "__dict__", ())):
            if name in self and self._is_mapped_attribute(name):
                yield name

    def keys(self, /) -> t.Iterable[t.Hashable]:
        """
        Yields the names of all items in the instance.
        """
        yield from self

    def values(self, /) -> t.Iterable[t.Any]:
        """
        Yields the values of all items in the instance.
        """
        for name in self:
            yield self[name]

    def items(self, /) -> t.Iterable[t.Tuple[t.Hashable, t.Any]]:
        """
        Yields the name, value pairs of all items in the instance.
        """
        for name in self:
            yield name, self[name]

    def update(
        self, mapping: t.Mapping[t.Hashable, t.Any] = {}, /, **kw: t.Any
    ) -> None:
        """
        Updates the instance with new key-values from a mapping.

        The defualt ``__init__`` relies on this method. This method
        shouldn't be overloaded in subclasses without care given to how
        it will change initialization as well.

        See: https://github.com/rmlibre/aiootp/pull/15
        """
        for name, value in {**dict(mapping), **kw}.items():
            self[name] = value


class OpenSlots(Slots):
    """
    A version of the `Slots` class which doesn't mask instance
    `__repr__`'s by default.

    Unmasked repr.
    Mutable instance.
    """

    __slots__ = ()

    def __repr__(self, /, *, mask: bool = False) -> str:
        """
        Denies setting attributes after they have already been set.
        """
        return super().__repr__(mask=mask)


class FrozenSlots(Slots):
    """
    A version of the `Slots` class which enables instances of subclasses
    to have attributes that are frozen once they're set.

    Masked repr.
    Immutable instance. (set once)
    """

    __slots__ = ()

    def __setitem__(self, name: str, value: t.Any, /) -> None:
        """
        Denies setting attributes after they have already been set.
        """
        if name in self:
            raise Issue.cant_reassign_attribute(name)

        super().__setitem__(name, value)

    def __delitem__(self, name: str, /) -> None:
        """
        Denies deleting attributes.
        """
        raise Issue.cant_deassign_attribute(name)


class OpenFrozenSlots(FrozenSlots):
    """
    A version of the `FrozenSlots` class which doesn't mask instance
    `__repr__`'s by default.

    Unmasked repr.
    Immutable instance. (set once)
    """

    __slots__ = ()

    def __repr__(self, /, *, mask: bool = False) -> str:
        """
        Denies setting attributes after they have already been set.
        """
        return super().__repr__(mask=mask)


module_api = dict(
    FrozenSlots=t.add_type(FrozenSlots),
    OpenFrozenSlots=t.add_type(OpenFrozenSlots),
    OpenSlots=t.add_type(OpenSlots),
    Slots=t.add_type(Slots),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Implementations & high-level interfaces to the package's online
salt misuse-reuse resistant, fully context committing, tweakable,
AEAD ciphers.
"""

__all__ = ["ChaCha20Poly1305", "Chunky2048", "Slick256"]


from cryptography.hazmat.primitives.ciphers.aead import ChaCha20Poly1305

from aiootp._typing import Typing as t

from .padding import *
from .ciphertext_formatting import *
from .cipher_kdfs import *
from .key_bundle import *
from .stream_hmac import *
from .synthetic_iv import *
from .stream_junction import *
from .cipher_stream_properties import *
from .cipher_streams import *
from .decipher_streams import *
from .dual_output_shake_cipher import *
from .dual_output_shake_cipher_config import *
from .shake_permute_cipher import *
from .shake_permute_cipher_config import *
from .cipher_interface import *
from .chunky_2048_config import *
from .chunky_2048 import *
from .slick_256_config import *
from .slick_256 import *


modules = dict(
    padding=padding,
    ciphertext_formatting=ciphertext_formatting,
    cipher_kdfs=cipher_kdfs,
    key_bundle=key_bundle,
    stream_hmac=stream_hmac,
    synthetic_iv=synthetic_iv,
    stream_junction=stream_junction,
    cipher_stream_properties=cipher_stream_properties,
    cipher_streams=cipher_streams,
    decipher_streams=decipher_streams,
    dual_output_shake_cipher=dual_output_shake_cipher,
    dual_output_shake_cipher_config=dual_output_shake_cipher_config,
    shake_permute_cipher=shake_permute_cipher,
    shake_permute_cipher_config=shake_permute_cipher_config,
    cipher_interface=cipher_interface,
    chunky_2048_config=chunky_2048_config,
    chunky_2048=chunky_2048,
    slick_256_config=slick_256_config,
    slick_256=slick_256,
)


module_api = dict(
    ChaCha20Poly1305=t.add_type(ChaCha20Poly1305),
    Chunky2048=Chunky2048,
    Ciphertext=Ciphertext,
    Padding=Padding,
    Slick256=Slick256,
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
A configuration type for the `AffineXORChain`.
"""

__all__ = ["AffineXORChainConfig"]


from aiootp._typing import Typing as t
from aiootp._exceptions import Issue
from aiootp.commons import Config


class AffineXORChainConfig(Config):
    """
    Configuration type for `AffineXORChain`.
    """

    __slots__ = (
        "SIZE",
        "IPAD",
        "OPAD",
        "XPAD",
        "DOMAIN",
        "KEY_TYPES",
        "KEY_SIZE",
        "ADD_KEY_SIZE",
        "XOR_KEY_SIZE",
        "PERMUTATION_CONFIG_ID",
        "Permutation",
    )

    slots_types: t.Mapping[str, type] = dict(
        SIZE=int,
        IPAD=int,
        OPAD=int,
        XPAD=int,
        DOMAIN=int,
        KEY_TYPES=tuple,
        KEY_SIZE=int,
        ADD_KEY_SIZE=int,
        XOR_KEY_SIZE=int,
        PERMUTATION_CONFIG_ID=t.Hashable,
        Permutation=t.PermutationType,
    )

    _DEFAULT_KEY_TYPES: t.Tuple[t.Tuple[str, int]] = (
        ("in_xor", 1),
        ("in_add", 2),
        ("mid_xor", 1),
        ("mid_add", 2),
        ("out_xor", 1),
        ("out_add", 2),
    )
    _FAST_CHAIN_KEY_TYPES: t.Tuple[t.Tuple[str, int]] = (
        ("add", 2),
        ("in_xor", 1),
        ("mid_xor", 1),
        ("out_xor", 1),
    )

    def _process_size(self, size: int) -> int:
        """
        Returns `size` if it passes preliminary non type-based value
        checks, otherwise raises `ValueError`.
        """
        if 4096 >= size > 0:
            return size
        else:
            raise Issue.value_must("size", "be > 0 and <= 4096")

    def _process_key_types(
        self, key_types: t.Optional[t.Tuple[t.Tuple[str, int]]]
    ) -> t.Tuple[t.Tuple[str, int]]:
        """
        Returns `key_types` if it contains at least one (str, int) pair,
        otherwise raises `ValueError`. Unless `key_types` isn't
        provided, then the class' default tuple is provided.
        """
        if key_types is None:
            return self._DEFAULT_KEY_TYPES
        elif not all(
            (
                name.isidentifier()
                and size_multiple.__class__ is int
                and 33 > size_multiple > 0
            )
            for name, size_multiple in key_types
        ):
            raise Issue.value_must("key_types", "contain (str, int) pairs")
        else:
            return key_types

    def _process_permutation_config_id(
        self, permutation_config_id: t.Optional[t.Hashable]
    ) -> t.Hashable:
        """
        Returns the input `permutation_config_id` if it's provided,
        otherwise defaults to returning the instance's `config_id`.
        """
        if permutation_config_id is None:
            return self.CONFIG_ID
        else:
            return permutation_config_id

    def _process_permutation(self, permutation_type: type) -> type:
        """
        Returns the input `permutation_type` if it follows the correct
        `PermutationType` protocol, otherwise raises `TypeError`.
        """
        if issubclass(permutation_type, t.PermutationType):
            return permutation_type
        else:
            raise Issue.value_must_be_subtype(
                f"{permutation_type=}", t.PermutationType
            )

    def __init__(
        self,
        *,
        config_id: t.Hashable,
        size: int,
        permutation_type: type,
        permutation_config_id: t.Optional[t.Hashable] = None,
        key_types: t.Optional[t.Tuple[t.Tuple[str, int]]] = None,
    ) -> None:
        """
        Caches the instance's static configuration after type-checking
        & doing injection of defaults for the inputs not provided.
        """
        self.CONFIG_ID = config_id
        self.SIZE = self._process_size(size)
        self.IPAD = int(self.SIZE * "a3", 16)
        self.OPAD = int(self.SIZE * "1b", 16)
        self.XPAD = int(self.SIZE * "8d", 16)
        self.DOMAIN = 1 << (8 * self.SIZE)
        self.KEY_TYPES = self._process_key_types(key_types)
        self.KEY_SIZE = self.SIZE * sum(
            multiple for name, multiple in self.KEY_TYPES
        )
        self.ADD_KEY_SIZE = 2 * self.SIZE
        self.XOR_KEY_SIZE = self.SIZE
        self.PERMUTATION_CONFIG_ID = self._process_permutation_config_id(
            permutation_config_id
        )
        self.Permutation = self._process_permutation(permutation_type)


module_api = dict(
    AffineXORChainConfig=t.add_type(AffineXORChainConfig),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Configuration logic & constants for `Slick256`.
"""

__all__ = ["slick256_spec"]


from hashlib import shake_128

from aiootp._typing import Typing as t
from aiootp._constants import EPOCH_NS, SECONDS
from aiootp._permutations import FastAffineXORChain

from .shake_permute_cipher_config import ShakePermuteCipherConfig
from .cipher_kdfs import ShakeConfig


BLOCK_ID_KDF: str = "block_id_kdf"
SHMAC_KDF: str = "shmac_kdf"


slick256_spec: t.ConfigType = ShakePermuteCipherConfig(
    name="Slick256",
    config_id=b"Slick256",
    blocksize=32,
    min_padding_blocks=0,
    epoch_ns=EPOCH_NS,
    time_unit=SECONDS,
    timestamp_bytes=4,
    siv_key_bytes=8,
    shmac_bytes=24,
    block_id_bytes=16,
    max_block_id_bytes=32,
    min_block_id_bytes=16,
    salt_bytes=8,
    iv_bytes=8,
    block_id_kdf=BLOCK_ID_KDF,
    block_id_kdf_config=ShakeConfig(
        name=BLOCK_ID_KDF,
        pad=b"\x9a",
        offset_amount=0,
        hasher=shake_128,
        key_slice=slice(None),
    ),
    shmac_kdf=SHMAC_KDF,
    shmac_kdf_config=ShakeConfig(
        name=SHMAC_KDF,
        pad=b"\xac",
        offset_amount=0,
        hasher=shake_128,
        key_slice=slice(None),
    ),
    permutation_type=FastAffineXORChain,
)


module_api = dict(
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    slick256_spec=slick256_spec,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Implementation of an Argon2i-like, memory hard, passphrase key
derivation function that's designed to be resistant to cache-timing
side-channel attacks & time-memory trade-offs.
"""

__all__ = ["Passcrypt"]


from .config import *
from .hash_format import *
from .session_init import *
from .sessions_manager import *
from .interface import *


modules = dict(
    config=config,
    hash_format=hash_format,
    session_init=session_init,
    sessions_manager=sessions_manager,
    interface=interface,
)


module_api = dict(
    PasscryptHash=PasscryptHash,
    Passcrypt=Passcrypt,
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Tools for time, asyncio, multi-threading, & multi-processing.
"""

__all__ = []


import asyncio
import aiofiles
from asyncio import run
from inspect import isawaitable as is_awaitable
from inspect import iscoroutinefunction as is_async_function

from .loops import *
from .clocks import *
from .aos import *
from .concurrency_interface import *
from .processes import *
from .threads import *


modules = dict(
    aos=aos,
    clocks=clocks,
    concurrency_interface=concurrency_interface,
    loops=loops,
    processes=processes,
    threads=threads,
)


module_api = dict(
    AsyncInit=AsyncInit,
    Clock=Clock,
    ConcurrencyGuard=ConcurrencyGuard,
    Processes=Processes,
    Threads=Threads,
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    aiofiles=aiofiles,
    asleep=asleep,
    asyncio=asyncio,
    gather=gather,
    is_async_function=is_async_function,
    is_awaitable=is_awaitable,
    get_event_loop_id=get_event_loop_id,
    get_process_id=get_process_id,
    get_thread_id=get_thread_id,
    new_future=new_future,
    new_task=new_task,
    run=run,
    sleep=sleep,
    wrap_in_executor=wrap_in_executor,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Configuration logic & constants for `Chunky2048`.
"""

__all__ = ["chunky2048_spec"]


from hashlib import shake_128

from aiootp._typing import Typing as t
from aiootp._constants import EPOCH_NS, SECONDS
from aiootp._permutations import FastAffineXORChain

from .dual_output_shake_cipher_config import DualOutputShakeCipherConfig
from .cipher_kdfs import ShakeConfig


BLOCK_ID_KDF: str = "block_id_kdf"
SHMAC_KDF: str = "shmac_kdf"
LEFT_KDF: str = "left_kdf"
RIGHT_KDF: str = "right_kdf"


chunky2048_spec: t.ConfigType = DualOutputShakeCipherConfig(
    name="Chunky2048",
    config_id=b"Chunky2048",
    blocksize=256,
    min_padding_blocks=0,
    epoch_ns=EPOCH_NS,
    time_unit=SECONDS,
    timestamp_bytes=4,
    siv_key_bytes=16,
    shmac_bytes=32,
    block_id_bytes=24,
    max_block_id_bytes=32,
    min_block_id_bytes=16,
    salt_bytes=8,
    iv_bytes=8,
    block_id_kdf=BLOCK_ID_KDF,
    block_id_kdf_config=ShakeConfig(
        name=BLOCK_ID_KDF,
        pad=b"\xac",
        offset_amount=0,
        hasher=shake_128,
        key_slice=slice(None),
    ),
    shmac_kdf=SHMAC_KDF,
    shmac_kdf_config=ShakeConfig(
        name=SHMAC_KDF,
        pad=b"\x9a",
        offset_amount=0,
        hasher=shake_128,
        key_slice=slice(None),
    ),
    left_kdf=LEFT_KDF,
    left_kdf_config=ShakeConfig(
        name=LEFT_KDF,
        pad=b"\x5c",
        offset_amount=0,
        hasher=shake_128,
        key_slice=slice(0, None, 2),  # Even index bytes
    ),
    right_kdf=RIGHT_KDF,
    right_kdf_config=ShakeConfig(
        name=RIGHT_KDF,
        pad=b"\x36",
        offset_amount=0,
        hasher=shake_128,
        key_slice=slice(1, None, 2),  # Odd index bytes
    ),
    permutation_type=FastAffineXORChain,
)


module_api = dict(
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    chunky2048_spec=chunky2048_spec,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Dependency inversion & documentation support for types relevant to
the `_permutations` subpackage.
"""

__all__ = ["PermutationType"]


from .interface import Typing as t


@t.runtime_checkable
class PermutationType(t.Protocol):
    @classmethod
    def key_size(cls, config_id: t.Hashable) -> int:
        pass  # pragma: no cover

    async def apermute(self, value: int) -> int:
        pass  # pragma: no cover

    def permute(self, value: int) -> int:
        pass  # pragma: no cover

    async def ainvert(self, value: int) -> int:
        pass  # pragma: no cover

    def invert(self, value: int) -> int:
        pass  # pragma: no cover


module_api = dict(
    PermutationType=t.add_type(PermutationType),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


"""
Time & timestamping tools.
"""

__all__ = [
    "Clock",
    "ns_counter",
    "s_counter",
    "this_nanosecond",
    "this_microsecond",
    "this_millisecond",
    "this_second",
    "this_minute",
    "this_hour",
    "this_day",
    "this_month",
    "this_year",
]


from time import time_ns, get_clock_info
from time import perf_counter as s_counter
from time import perf_counter_ns as ns_counter

from aiootp._typing import Typing as t
from aiootp._constants import SECONDS, EPOCH_NS, BIG
from aiootp._constants import SAFE_TIMESTAMP_BYTES
from aiootp._exceptions import Issue, TimestampExpired
from aiootp.commons import FrozenNamespace, FrozenInstance

from .loops import asleep


_YEAR_WITH_LEAP_DAYS: float = 365.24225
_ONE_NANOSECOND: int = 1
_ONE_MICROSECOND: int = 1_000
_ONE_MILLISECOND: int = 1_000_000
_ONE_SECOND: int = 1_000_000_000
_ONE_MINUTE: int = 60 * _ONE_SECOND
_ONE_HOUR: int = 60 * _ONE_MINUTE
_ONE_DAY: int = 24 * _ONE_HOUR
_ONE_MONTH: float = (_YEAR_WITH_LEAP_DAYS * _ONE_DAY) / 12
_ONE_YEAR: float = _YEAR_WITH_LEAP_DAYS * _ONE_DAY


def this_nanosecond(epoch: int = 0) -> int:
    return time_ns() - epoch


def this_microsecond(epoch: int = 0) -> int:
    return this_nanosecond(epoch) // _ONE_MICROSECOND


def this_millisecond(epoch: int = 0) -> int:
    return this_nanosecond(epoch) // _ONE_MILLISECOND


def this_second(epoch: int = 0) -> int:
    return this_nanosecond(epoch) // _ONE_SECOND


def this_minute(epoch: int = 0) -> int:
    return this_nanosecond(epoch) // _ONE_MINUTE


def this_hour(epoch: int = 0) -> int:
    return this_nanosecond(epoch) // _ONE_HOUR


def this_day(epoch: int = 0) -> int:
    return this_nanosecond(epoch) // _ONE_DAY


def this_month(epoch: int = 0) -> int:
    return int(this_nanosecond(epoch) / _ONE_MONTH)


def this_year(epoch: int = 0) -> int:
    return int(this_nanosecond(epoch) / _ONE_YEAR)


class Clock(FrozenInstance):
    """
    A class whose objects are used for creating & measuring bytes-type
    timestamps, with configurable time units & epoch of measure.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    from aiootp.asynchs import Clock

    # Create an object with specified units & epoch ->
    ns_clock = Clock("nanoseconds", epoch=0)

    # Create a bytes-type timestamp of the object's current time in
    # nanoseconds from its epoch ->
    timestamp = ns_clock.make_timestamp(size=8)

    # Retrieve the elapsed time from the object's current time & a given
    # timestamp ->
    while ns_clock.delta(timestamp) < 2_000_000_000: # wait two seconds
        await do_something_else()

    # Throw a `TimestampExpired` error if a given timestamp is older
    # than `ttl` units from the object's current time ->
    try:
        ns_clock.test_timestamp(timestamp, ttl=1_000_000_000)
    except ns_clock.TimestampExpired as e:
        print(f"Timestamp expired by {e.expired_by} # of {e.units}.")
        'Timestamp expired by 287491003983 # of nanoseconds.'

    # These are the supported units ->
    year_clock = Clock("years")
    month_clock = Clock("months")
    day_clock = Clock("days")
    hour_clock = Clock("hours")
    minute_clock = Clock("minutes")
    second_clock = Clock("seconds")
    ms_clock = Clock("milliseconds")
    µs_clock = Clock("microseconds")
    ns_clock = Clock("nanoseconds")

    # The `epoch` is always measured in nanoseconds from the UNIX epoch of 0
    hour_clock = Clock("hours", epoch=9000)  # time starts 9000 nanoseconds
                                             # after the UNIX epoch
    # The default epoch for the package is 1672531200000000000,
    # Sun, 01 Jan 2023 00:00:00 UTC
    """

    __slots__ = ("_epoch", "_resolution_needed", "_time", "_units")

    _SYSTEM_TIME_RESOLUTION: t.PositiveRealNumber = get_clock_info(
        "time"
    ).resolution

    _times: FrozenNamespace = FrozenNamespace(
        years=(this_year, _ONE_YEAR / _ONE_SECOND),
        months=(this_month, _ONE_MONTH / _ONE_SECOND),
        days=(this_day, 24 * 60 * 60),
        hours=(this_hour, 60 * 60),
        minutes=(this_minute, 60),
        seconds=(this_second, 1),
        milliseconds=(this_millisecond, 1e-3),
        microseconds=(this_microsecond, 1e-6),
        nanoseconds=(this_nanosecond, 1e-9),
    )

    TimestampExpired: type = TimestampExpired

    def __init__(
        self, /, units: str = SECONDS, *, epoch: int = EPOCH_NS
    ) -> None:
        """
        Create an object which can create & measure bytes-type
        timestamps, with configurable units & epoch of measure.
        """
        if units not in self._times:
            raise Issue.invalid_value("time units", units)
        self._epoch = epoch
        self._time, self._resolution_needed = self._times[units]
        self._units = units

    def __repr__(self, /) -> str:
        return (
            f"{self.__class__.__qualname__}("
            f"{self._units!r}, epoch={self._epoch})"
        )

    def has_adequate_resolution(self, /) -> bool:
        """
        Reports on whether the system's time resolution in fine enough
        to accurately work with the requested time units.
        """
        return self._resolution_needed >= self._SYSTEM_TIME_RESOLUTION

    async def atime(self, /) -> int:
        """
        Returns the instance's conception of the current time as an
        integer, which is the number of time units since the instance's
        epoch.
        """
        await asleep()
        return self._time(self._epoch)

    def time(self, /) -> int:
        """
        Returns the instance's conception of the current time as an
        integer, which is the number of time units since the instance's
        epoch.
        """
        return self._time(self._epoch)

    async def amake_timestamp(
        self,
        /,
        *,
        size: int = SAFE_TIMESTAMP_BYTES,
        byte_order: str = BIG,
    ) -> bytes:
        """
        Returns a `size`-byte `byte_order`-endian representation of
        the instance's conception of the current time.
        """
        return (await self.atime()).to_bytes(size, byte_order)

    def make_timestamp(
        self,
        /,
        *,
        size: int = SAFE_TIMESTAMP_BYTES,
        byte_order: str = BIG,
    ) -> bytes:
        """
        Returns a `size`-byte `byte_order`-endian representation of
        the instance's conception of the current time.
        """
        return self.time().to_bytes(size, byte_order)

    async def aread_timestamp(
        self, /, timestamp: bytes, *, byte_order: str = BIG
    ) -> int:
        """
        Returns the integer representation of the `byte_order`-endian
        bytes-type `timestamp`.
        """
        await asleep()
        return int.from_bytes(timestamp, byte_order)

    def read_timestamp(
        self, /, timestamp: bytes, *, byte_order: str = BIG
    ) -> int:
        """
        Returns the integer representation of the `byte_order`-endian
        bytes-type `timestamp`.
        """
        return int.from_bytes(timestamp, byte_order)

    async def adelta(
        self, /, timestamp: bytes, *, byte_order: str = BIG
    ) -> int:
        """
        Takes a `timestamp` & returns the integer difference between
        the instance's conception of the current time & the timestamp.
        """
        return await self.atime() - await self.aread_timestamp(
            timestamp, byte_order=byte_order
        )

    def delta(self, /, timestamp: bytes, *, byte_order: str = BIG) -> int:
        """
        Takes a `timestamp` & returns the integer difference between
        the instance's conception of the current time & the timestamp.
        """
        return self.time() - self.read_timestamp(
            timestamp, byte_order=byte_order
        )

    async def atest_timestamp(
        self,
        /,
        timestamp: bytes,
        ttl: t.Optional[int],
        *,
        byte_order: str = BIG,
    ) -> None:
        """
        Raises `TimestampExpired` if `timestamp` is more than
        `ttl` time units old from the instance's conception of the
        current time.
        """
        if ttl is None:
            return
        delta = await self.adelta(timestamp, byte_order=byte_order)
        timestamp_is_expired = delta > ttl
        expired_by = delta - ttl
        if timestamp_is_expired:
            raise self.TimestampExpired(self._units, expired_by)

    def test_timestamp(
        self,
        /,
        timestamp: bytes,
        ttl: t.Optional[int],
        *,
        byte_order: str = BIG,
    ) -> None:
        """
        Raises `TimestampExpired` if `timestamp` is more than
        `ttl` time units old from the instance's conception of the
        current time.
        """
        if ttl is None:
            return
        delta = self.delta(timestamp, byte_order=byte_order)
        timestamp_is_expired = delta > ttl
        expired_by = delta - ttl
        if timestamp_is_expired:
            raise self.TimestampExpired(self._units, expired_by)


module_api = dict(
    Clock=t.add_type(Clock),
    __all__=__all__,
    __doc__=__doc__,
    __file__=__file__,
    __name__=__name__,
    __spec__=__spec__,
    __loader__=__loader__,
    __package__=__package__,
    ns_counter=ns_counter,
    s_counter=s_counter,
    this_nanosecond=this_nanosecond,
    this_microsecond=this_microsecond,
    this_millisecond=this_millisecond,
    this_second=this_second,
    this_minute=this_minute,
    this_hour=this_hour,
    this_day=this_day,
    this_month=this_month,
    this_year=this_year,
)



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


import math
import time
import warnings

from conftest import *


TIME_RESOLUTION = time.get_clock_info("time").resolution
TIME_VARIANCE = math.ceil(TIME_RESOLUTION / 1e-09)

YEAR_WITH_LEAP_DAYS = 365.24225

EPOCHS_TESTED = (
    OpenNamespace(seconds=0, nanoseconds=0),
    OpenNamespace(seconds=EPOCH, nanoseconds=EPOCH_NS),
)


class EqualTimingExperiment:
    __slots__ = ("early_control", "experiment", "late_control", "epoch")

    def __init__(
        self,
        control_timer: t.Callable[[], t.PositiveRealNumber],
        experiment_timer: t.Callable[[], int],
        epoch: int,
    ) -> None:
        self.epoch = epoch
        self.early_control, self.experiment, self.late_control = (
            control_timer(),
            experiment_timer(),
            control_timer(),
        )

    def correct_range(
        self,
        *,
        control_conversion: t.Callable[[t.PositiveRealNumber], int],
    ) -> range:
        return range(
            control_conversion(self.early_control, epoch=self.epoch) - 1,
            control_conversion(self.late_control, epoch=self.epoch) + 2,
        )


class TestAPlatformCounter:
    def test_is_monotonic(self) -> None:
        assert time.get_clock_info("perf_counter").monotonic is True

    def test_is_nanosecond_precise(self) -> None:
        problem = (  # fmt: skip
            "Platform perf counter doesn't have nanosecond resolution."
        )
        resolution_warning = lambda _: warnings.warn(problem) or True
        with Ignore(AssertionError, if_except=resolution_warning):
            assert time.get_clock_info("perf_counter").resolution <= 1e-09
        assert time.get_clock_info("perf_counter").resolution <= 1e-07


class TestAPlatformTime:
    def test_is_at_least_millisecond_precise(self) -> None:
        problem = (  # fmt: skip
            "Platform time doesn't have at least millisecond resolution."
        )
        resolution_warning = lambda _: warnings.warn(problem) or True
        with Ignore(AssertionError, if_except=resolution_warning):
            assert time.get_clock_info("time").resolution <= 1e-03
        assert time.get_clock_info("time").resolution <= 1 / 64


class TestClockConversions:
    seconds_to_nanoseconds = staticmethod(
        lambda control, epoch: int(
            (control - epoch.seconds) * 1_000_000_000
        )
    )
    seconds_to_microseconds = staticmethod(
        lambda control, epoch: int((control - epoch.seconds) * 1_000_000)
    )
    seconds_to_milliseconds = staticmethod(
        lambda control, epoch: int((control - epoch.seconds) * 1_000)
    )
    seconds_to_seconds = staticmethod(
        lambda control, epoch: int(control - epoch.seconds)
    )
    seconds_to_minutes = staticmethod(
        lambda control, epoch: int((control - epoch.seconds) / 60)
    )
    seconds_to_hours = staticmethod(
        lambda control, epoch: int((control - epoch.seconds) / (60 * 60))
    )
    seconds_to_days = staticmethod(
        lambda control, epoch: int(
            (control - epoch.seconds) / (60 * 60 * 24)
        )
    )
    seconds_to_months = staticmethod(
        lambda control, epoch: int(
            (control - epoch.seconds)
            / (60 * 60 * 24 * YEAR_WITH_LEAP_DAYS / 12)
        )
    )
    seconds_to_years = staticmethod(
        lambda control, epoch: int(
            (control - epoch.seconds) / (60 * 60 * 24 * YEAR_WITH_LEAP_DAYS)
        )
    )

    def test_package_epoch_starts_year_2023(self) -> None:
        info = time.gmtime(Clock(SECONDS, epoch=0).time())
        year_difference = info.tm_year - 2023
        assert year_difference >= 0
        package_info = time.gmtime(Clock(SECONDS).time())
        assert package_info.tm_year - 1970 == year_difference

    def test_epoch_representations_are_equivalent(self) -> None:
        assert EPOCH == (EPOCH_NS // 1_000_000_000)

    def test_nanoseconds_are_always_unique_and_incrementing(self) -> None:
        number_of_tests = 1024
        iterations = tuple(range(number_of_tests))
        right_now = Clock(NANOSECONDS).time
        tests = [right_now() for test in iterations]
        if TIME_RESOLUTION <= 1e-07:
            assert number_of_tests == len(set(tests))
        elif TIME_RESOLUTION <= 1e-03:
            assert number_of_tests <= 10 * len(set(tests))
        assert tests == sorted(tests)

    def test_nanoseconds_correctness(self) -> None:
        for epoch in EPOCHS_TESTED:
            clock = Clock(NANOSECONDS, epoch=epoch.nanoseconds)
            test = EqualTimingExperiment(
                control_timer=time.time,
                experiment_timer=clock.time,
                epoch=epoch,
            )
            span = test.correct_range(
                control_conversion=self.seconds_to_nanoseconds
            )
            if TIME_RESOLUTION <= 1e-09:
                assert test.experiment in span
            else:
                assert test.experiment in range(
                    span.start - TIME_VARIANCE,
                    span.stop + TIME_VARIANCE,
                )

    def test_microseconds_correctness(self) -> None:
        for epoch in EPOCHS_TESTED:
            clock = Clock(MICROSECONDS, epoch=epoch.nanoseconds)
            test = EqualTimingExperiment(
                control_timer=time.time,
                experiment_timer=clock.time,
                epoch=epoch,
            )
            span = test.correct_range(
                control_conversion=self.seconds_to_microseconds
            )
            if TIME_RESOLUTION <= 1e-09:
                assert test.experiment in span
            else:
                assert test.experiment in range(
                    span.start - (TIME_VARIANCE // 1_000),
                    span.stop + (TIME_VARIANCE // 1_000),
                )

    def test_milliseconds_correctness(self) -> None:
        for epoch in EPOCHS_TESTED:
            clock = Clock(MILLISECONDS, epoch=epoch.nanoseconds)
            test = EqualTimingExperiment(
                control_timer=time.time,
                experiment_timer=clock.time,
                epoch=epoch,
            )
            span = test.correct_range(
                control_conversion=self.seconds_to_milliseconds
            )
            if TIME_RESOLUTION <= 1e-09:
                assert test.experiment in span
            else:
                assert test.experiment in range(
                    span.start - (TIME_VARIANCE // 1_000_000),
                    span.stop + (TIME_VARIANCE // 1_000_000),
                )

    def test_seconds_correctness(self) -> None:
        for epoch in EPOCHS_TESTED:
            clock = Clock(SECONDS, epoch=epoch.nanoseconds)
            test = EqualTimingExperiment(
                control_timer=time.time,
                experiment_timer=clock.time,
                epoch=epoch,
            )
            assert test.experiment in test.correct_range(
                control_conversion=self.seconds_to_seconds
            )

    def test_minutes_correctness(self) -> None:
        for epoch in EPOCHS_TESTED:
            clock = Clock(MINUTES, epoch=epoch.nanoseconds)
            test = EqualTimingExperiment(
                control_timer=time.time,
                experiment_timer=clock.time,
                epoch=epoch,
            )
            assert test.experiment in test.correct_range(
                control_conversion=self.seconds_to_minutes
            )

    def test_hours_correctness(self) -> None:
        for epoch in EPOCHS_TESTED:
            clock = Clock(HOURS, epoch=epoch.nanoseconds)
            test = EqualTimingExperiment(
                control_timer=time.time,
                experiment_timer=clock.time,
                epoch=epoch,
            )
            assert test.experiment in test.correct_range(
                control_conversion=self.seconds_to_hours
            )

    def test_days_correctness(self) -> None:
        for epoch in EPOCHS_TESTED:
            clock = Clock(DAYS, epoch=epoch.nanoseconds)
            test = EqualTimingExperiment(
                control_timer=time.time,
                experiment_timer=clock.time,
                epoch=epoch,
            )
            assert test.experiment in test.correct_range(
                control_conversion=self.seconds_to_days
            )

    def test_months_correctness(self) -> None:
        for epoch in EPOCHS_TESTED:
            clock = Clock(MONTHS, epoch=epoch.nanoseconds)
            test = EqualTimingExperiment(
                control_timer=time.time,
                experiment_timer=clock.time,
                epoch=epoch,
            )
            assert test.experiment in test.correct_range(
                control_conversion=self.seconds_to_months
            )

    def test_years_correctness(self) -> None:
        for epoch in EPOCHS_TESTED:
            clock = Clock(YEARS, epoch=epoch.nanoseconds)
            test = EqualTimingExperiment(
                control_timer=time.time,
                experiment_timer=clock.time,
                epoch=epoch,
            )
            assert test.experiment in test.correct_range(
                control_conversion=self.seconds_to_years
            )


class TestClock:
    async def test_invalid_unit_throws_error(self) -> None:
        problem = (  # fmt: skip
            "An invalid time unit successfully initialized an instance."
        )
        with Ignore(ValueError, if_else=violation(problem)):
            Clock("attoseconds")

    async def test_instance_repr(self) -> None:
        units = SECONDS
        epoch = EPOCH_NS
        string = repr(Clock(units, epoch=epoch))
        assert str(units) in string
        assert str(epoch) in string


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})



# This file is part of aiootp:
# a high-level async cryptographic anonymity library to scale, simplify,
# & automate privacy best practices for secure data & identity processing,
# communication, & storage.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2024 Ricchi (Richard) Machado <rmlibre@riseup.net>
# All rights reserved.
#


from conftest import *


async def test_json_functions() -> None:
    for _, cipher, salt, aad in all_ciphers:
        ciphertext_of_dict = cipher.json_encrypt(test_data, salt=salt)
        ciphertext_of_string = cipher.json_encrypt(
            plaintext_string, salt=salt
        )

        # plaintext is not trivially detected in ciphertext
        assert json.dumps(test_data).encode() not in ciphertext_of_dict
        assert string_leakage not in ciphertext_of_string

        # no two ciphertexts are ever the same
        assert ciphertext_of_dict != await cipher.ajson_encrypt(
            test_data, salt=salt
        )
        assert ciphertext_of_string != await cipher.ajson_encrypt(
            plaintext_string, salt=salt
        )

        assert ciphertext_of_dict != cipher.json_encrypt(test_data)
        assert ciphertext_of_string != await cipher.ajson_encrypt(
            plaintext_string
        )

        assert ciphertext_of_dict != cipher.json_encrypt(
            test_data, salt=salt, aad=aad
        )
        assert ciphertext_of_string != await cipher.ajson_encrypt(
            plaintext_string, salt=salt, aad=aad
        )

        # decryption of correct ciphertext & key of json data doesn't fail
        # & is sound
        plaintext_of_dict = cipher.json_decrypt(ciphertext_of_dict)
        plaintext_of_string = cipher.json_decrypt(ciphertext_of_string)
        assert test_data == plaintext_of_dict
        assert plaintext_string == plaintext_of_string

        # async decryption with correct ciphertext & key of json data
        # doesn't fail & is sound
        assert plaintext_of_dict == await cipher.ajson_decrypt(
            ciphertext_of_dict
        )
        assert plaintext_of_string == await cipher.ajson_decrypt(
            ciphertext_of_string
        )


async def test_bytes_functions() -> None:
    for _, cipher, salt, aad in all_ciphers:
        ciphertext_of_bytes = cipher.bytes_encrypt(
            plaintext_bytes, salt=salt
        )
        ciphertext_of_string = cipher.bytes_encrypt(
            plaintext_string.encode(), salt=salt
        )

        # plaintext is not trivially detected in ciphertext
        assert byte_leakage not in ciphertext_of_bytes
        assert string_leakage not in ciphertext_of_string

        # no two ciphertexts are ever the same
        assert ciphertext_of_bytes != await cipher.abytes_encrypt(
            plaintext_bytes, salt=salt
        )
        assert ciphertext_of_string != await cipher.abytes_encrypt(
            plaintext_string.encode(), salt=salt
        )

        assert ciphertext_of_bytes != cipher.bytes_encrypt(plaintext_bytes)
        assert ciphertext_of_string != await cipher.abytes_encrypt(
            plaintext_string.encode()
        )

        assert ciphertext_of_bytes != cipher.bytes_encrypt(
            plaintext_bytes, salt=salt, aad=aad
        )
        assert ciphertext_of_string != await cipher.abytes_encrypt(
            plaintext_string.encode(), salt=salt, aad=aad
        )

        # decryption of correct ciphertext & key doesn't fail & is sound
        plaintext_of_bytes = cipher.bytes_decrypt(ciphertext_of_bytes)
        plaintext_of_string = cipher.bytes_decrypt(ciphertext_of_string)
        assert plaintext_bytes == plaintext_of_bytes
        assert plaintext_string == plaintext_of_string.decode()

        # async decryption with correct ciphertext & key doesn't fail & is sound
        assert plaintext_of_bytes == await cipher.abytes_decrypt(
            ciphertext_of_bytes
        )
        assert plaintext_of_string == await cipher.abytes_decrypt(
            ciphertext_of_string
        )


async def test_token_functions() -> None:
    for _, cipher, _, aad in all_ciphers:
        ciphertext_of_bytes = cipher.make_token(plaintext_bytes, aad=aad)

        # no two ciphertexts are ever the same
        problem = (  # fmt: skip
            "A non-bytes plaintext was allowed."
        )
        with Ignore(TypeError, if_else=violation(problem)):
            assert await cipher.amake_token("plaintext", aad=aad)
        with Ignore(TypeError, if_else=violation(problem)):
            assert cipher.make_token("plaintext", aad=aad)

        assert ciphertext_of_bytes != cipher.make_token(
            plaintext_bytes, aad=aad
        )
        assert ciphertext_of_bytes != await cipher.amake_token(
            plaintext_bytes, aad=aad
        )
        assert ciphertext_of_bytes != cipher.make_token(plaintext_bytes)
        assert ciphertext_of_bytes != await cipher.amake_token(
            plaintext_bytes
        )

        # token decryption of correct ciphertext & key doesn't fail & is sound
        assert plaintext_bytes == cipher.read_token(
            ciphertext_of_bytes, aad=aad
        )
        assert plaintext_bytes == cipher.read_token(
            ciphertext_of_bytes.decode(), aad=aad
        )
        assert plaintext_bytes == await cipher.aread_token(
            ciphertext_of_bytes, aad=aad
        )
        assert plaintext_bytes == await cipher.aread_token(
            ciphertext_of_bytes.decode(), aad=aad
        )


async def test_chunky2048_test_vectors() -> None:
    test_vectors = [chunky2048_test_vector_0]

    for test in test_vectors:
        cipher = Chunky2048(test.key)
        assert test.plaintext == cipher.bytes_decrypt(
            test.shmac + test.salt + test.iv + test.ciphertext,
            aad=test.aad,
        )


async def test_slick256_test_vectors() -> None:
    test_vectors = [slick256_test_vector_0]

    for test in test_vectors:
        cipher = Slick256(test.key)
        assert test.plaintext == cipher.bytes_decrypt(
            test.shmac + test.salt + test.iv + test.ciphertext,
            aad=test.aad,
        )


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})
